import textwrap
from os.path import abspath, dirname, join

import pytest

from astropy import extern
from astropy.coordinates import SkyCoord
from astropy.table.table import Table
from astropy.time import Time
from astropy.utils.compat.optional_deps import (
    HAS_BLEACH,
    HAS_IPYDATAGRID,
    HAS_IPYTHON,
    HAS_PANDAS,
)
from astropy.utils.exceptions import AstropyDeprecationWarning
from astropy.utils.misc import _NOT_OVERWRITING_MSG_MATCH

EXTERN_DIR = abspath(join(dirname(extern.__file__), "jquery", "data"))
JQUERY_MIN_JS = "jquery-3.6.0.min.js"


REFERENCE = """
<html>
 <head>
  <meta charset="utf-8"/>
  <meta content="text/html;charset=UTF-8" http-equiv="Content-type"/>
  <style>
body {font-family: sans-serif;}
table.dataTable {width: auto !important; margin: 0 !important;}
.dataTables_filter, .dataTables_paginate {float: left !important; margin-left:1em}
  </style>
  <link href="%(datatables_css_url)s" rel="stylesheet" type="text/css"/>
  <script src="%(jquery_url)s">
  </script>
  <script src="%(datatables_js_url)s">
  </script>
 </head>
 <body>
  <script>
var astropy_sort_num = function(a, b) {
    var a_num = parseFloat(a);
    var b_num = parseFloat(b);

    if (isNaN(a_num) && isNaN(b_num))
        return ((a < b) ? -1 : ((a > b) ? 1 : 0));
    else if (!isNaN(a_num) && !isNaN(b_num))
        return ((a_num < b_num) ? -1 : ((a_num > b_num) ? 1 : 0));
    else
        return isNaN(a_num) ? -1 : 1;
}

jQuery.extend( jQuery.fn.dataTableExt.oSort, {
    "optionalnum-asc": astropy_sort_num,
    "optionalnum-desc": function (a,b) { return -astropy_sort_num(a, b); }
});

$(document).ready(function() {
    $('#%(table_id)s').dataTable({
        order: [],
        pageLength: %(length)s,
        lengthMenu: [[%(display_length)s, -1], [%(display_length)s, 'All']],
        pagingType: "full_numbers",
        columnDefs: [{targets: [0], type: "optionalnum"}]
    });
} );  </script>
  <table class="%(table_class)s" id="%(table_id)s">
   <thead>
    <tr>
     <th>a</th>
     <th>b</th>
    </tr>
   </thead>
%(lines)s
  </table>
 </body>
</html>
"""

TPL = "   <tr>\n    <td>{0}</td>\n    <td>{1}</td>\n   </tr>"


def format_lines(col1, col2):
    col1_format = getattr(col1.info, "default_format", lambda x: x)
    col2_format = getattr(col2.info, "default_format", lambda x: x)
    return "\n".join(
        TPL.format(col1_format(v1), col2_format(v2)) for v1, v2 in zip(col1, col2)
    )


def test_write_jsviewer_default(tmp_path):
    t = Table()
    t["a"] = [1, 2, 3, 4, 5]
    t["b"] = ["a", "b", "c", "d", "e"]
    t["a"].unit = "m"

    tmpfile = tmp_path / "test.html"

    t.write(tmpfile, format="jsviewer")
    ref = REFERENCE % dict(
        lines=format_lines(t["a"], t["b"]),
        table_class="display compact",
        table_id=f"table{id(t)}",
        length="50",
        display_length="10, 25, 50, 100, 500, 1000",
        datatables_css_url=(
            "https://cdn.datatables.net/2.1.8/css/dataTables.dataTables.min.css"
        ),
        datatables_js_url=("https://cdn.datatables.net/2.1.8/js/dataTables.min.js"),
        jquery_url="https://code.jquery.com/" + JQUERY_MIN_JS,
    )
    with open(tmpfile) as f:
        assert f.read().strip() == ref.strip()


def test_write_jsviewer_overwrite(tmp_path):
    t = Table()
    t["a"] = [1, 2, 3, 4, 5]
    t["b"] = ["a", "b", "c", "d", "e"]
    t["a"].unit = "m"
    tmpfile = tmp_path / "test.html"

    # normal write
    t.write(tmpfile, format="jsviewer")
    # errors on overwrite
    with pytest.raises(OSError, match=_NOT_OVERWRITING_MSG_MATCH):
        t.write(tmpfile, format="jsviewer")
    # unless specified
    t.write(tmpfile, format="jsviewer", overwrite=True)


@pytest.mark.parametrize(
    "mixin",
    [
        Time(["J2000", "J2001"]),
        Time([50000.0, 50001.0001], format="mjd"),
        SkyCoord(ra=[100.0, 110.0], dec=[-10.0, 10.0], unit="deg"),
    ],
)
def test_write_jsviewer_mixin(tmp_path, mixin):
    t = Table()
    t["a"] = [1, 2]
    t["b"] = mixin
    t["a"].unit = "m"

    tmpfile = tmp_path / "test.html"

    t.write(tmpfile, format="jsviewer")
    ref = REFERENCE % dict(
        lines=format_lines(t["a"], t["b"]),
        table_class="display compact",
        table_id=f"table{id(t)}",
        length="50",
        display_length="10, 25, 50, 100, 500, 1000",
        datatables_css_url=(
            "https://cdn.datatables.net/2.1.8/css/dataTables.dataTables.min.css"
        ),
        datatables_js_url=("https://cdn.datatables.net/2.1.8/js/dataTables.min.js"),
        jquery_url="https://code.jquery.com/" + JQUERY_MIN_JS,
    )
    with open(tmpfile) as f:
        assert f.read().strip() == ref.strip()


@pytest.mark.skipif(not HAS_BLEACH, reason="requires bleach")
def test_write_jsviewer_options(tmp_path):
    t = Table()
    t["a"] = [1, 2, 3, 4, 5]
    t["b"] = ["<b>a</b>", "b", "c", "d", "e"]
    t["a"].unit = "m"

    tmpfile = tmp_path / "test.html"
    t.write(
        tmpfile,
        format="jsviewer",
        table_id="test",
        max_lines=3,
        jskwargs={"display_length": 5},
        table_class="display hover",
        htmldict=dict(raw_html_cols="b"),
    )

    ref = REFERENCE % dict(
        lines=format_lines(t["a"][:3], t["b"][:3]),
        table_class="display hover",
        table_id="test",
        length="5",
        display_length="5, 10, 25, 50, 100, 500, 1000",
        datatables_css_url=(
            "https://cdn.datatables.net/2.1.8/css/dataTables.dataTables.min.css"
        ),
        datatables_js_url=("https://cdn.datatables.net/2.1.8/js/dataTables.min.js"),
        jquery_url="https://code.jquery.com/" + JQUERY_MIN_JS,
    )
    with open(tmpfile) as f:
        assert f.read().strip() == ref.strip()


@pytest.mark.skipif(not HAS_IPYTHON, reason="requires IPython")
def test_show_in_notebook_classic():
    t = Table()
    t["a"] = [1, 2, 3, 4, 5]
    t["b"] = ["b", "c", "a", "d", "e"]

    with pytest.warns(AstropyDeprecationWarning):
        htmlstr_windx = t.show_in_notebook(
            backend="classic"
        ).data  # should default to 'idx'
        htmlstr_windx_named = t.show_in_notebook(
            backend="classic", show_row_index="realidx"
        ).data
        htmlstr_woindx = t.show_in_notebook(
            backend="classic", show_row_index=False
        ).data

    assert (
        textwrap.dedent(
            """
    <thead><tr><th>idx</th><th>a</th><th>b</th></tr></thead>
    <tr><td>0</td><td>1</td><td>b</td></tr>
    <tr><td>1</td><td>2</td><td>c</td></tr>
    <tr><td>2</td><td>3</td><td>a</td></tr>
    <tr><td>3</td><td>4</td><td>d</td></tr>
    <tr><td>4</td><td>5</td><td>e</td></tr>
    """
        ).strip()
        in htmlstr_windx
    )

    assert (
        "<thead><tr><th>realidx</th><th>a</th><th>b</th></tr></thead>"
        in htmlstr_windx_named
    )

    assert "<thead><tr><th>a</th><th>b</th></tr></thead>" in htmlstr_woindx


@pytest.mark.skipif(
    not HAS_IPYDATAGRID or not HAS_PANDAS, reason="requires ipydatagrid and pandas"
)
# https://github.com/bqplot/bqplot/issues/1624 and such
@pytest.mark.filterwarnings(r"ignore:((.|\n)*)traitlets((.|\n)*):DeprecationWarning")
def test_show_in_notebook_ipydatagrid():
    from ipydatagrid import DataGrid

    t = Table()
    dg = t.show_in_notebook()
    assert isinstance(dg, DataGrid)


def test_show_in_notebook_invalid_backend():
    t = Table()
    with pytest.raises(NotImplementedError, match=".* backend is not supported"):
        t.show_in_notebook(backend="foo")



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Handles the "FITS" unit format.
"""

from typing import Literal

import numpy as np

from astropy.units.core import CompositeUnit, UnitBase
from astropy.units.enums import DeprecatedUnitAction
from astropy.units.errors import UnitScaleError
from astropy.utils import classproperty

from . import Base
from .generic import _GenericParserMixin


class FITS(Base, _GenericParserMixin):
    """
    The FITS standard unit format.

    This supports the format defined in the Units section of the `FITS
    Standard <https://fits.gsfc.nasa.gov/fits_standard.html>`_.
    """

    @classproperty(lazy=True)
    def _units(cls) -> dict[str, UnitBase]:
        from astropy import units as u

        names = {
            "Celsius": u.deg_C,
            "deg C": u.deg_C,
            "dbyte": u.Unit("dbyte", 0.1 * u.byte),
            "as": u.attosecond,
        }
        bases = [
            "m", "g", "s", "rad", "sr", "K", "A", "mol", "cd",
            "Hz", "J", "W", "V", "N", "Pa", "C", "Ohm", "S",
            "F", "Wb", "T", "H", "lm", "lx", "a", "yr", "eV",
            "pc", "Jy", "mag", "R", "bit", "byte", "G", "barn",
        ]  # fmt: skip
        prefixes = [
            "y", "z", "a", "f", "p", "n", "u", "m", "c", "d",
            "", "da", "h", "k", "M", "G", "T", "P", "E", "Z", "Y",
        ]  # fmt: skip

        for key in (prefix + base for base in bases for prefix in prefixes):
            if key not in names:
                names[key] = getattr(u, key)
        simple_units = [
            "deg", "arcmin", "arcsec", "mas", "min", "h", "d", "Ry",
            "solMass", "u", "solLum", "solRad", "AU", "lyr", "count",
            "ct", "photon", "ph", "pixel", "pix", "D", "Sun", "chan",
            "bin", "voxel", "adu", "beam", "erg", "Angstrom", "angstrom",
        ]  # fmt: skip
        names.update((unit, getattr(u, unit)) for unit in simple_units)

        return names

    @classmethod
    def to_string(
        cls,
        unit: UnitBase,
        fraction: bool | Literal["inline", "multiline"] = False,
        deprecations: DeprecatedUnitAction = DeprecatedUnitAction.WARN,
    ) -> str:
        # Remove units that aren't known to the format
        unit = cls._decompose_to_known_units(unit)

        parts = []

        base = np.log10(unit.scale)

        if base % 1.0 != 0.0:
            raise UnitScaleError(
                "The FITS unit format is not able to represent scales "
                "that are not powers of 10.  Multiply your data by "
                f"{unit.scale:e}."
            )
        elif unit.scale != 1.0:
            # We could override format_exponential_notation to set the
            # scale factor but that would give the wrong impression that
            # all values in FITS are set that way.  So, instead do it
            # here, and use a unity-scale unit for the rest.
            parts.append(f"10**{int(base)}")
            unit = CompositeUnit(1, unit.bases, unit.powers)

        if unit.bases:
            parts.append(super().to_string(unit, fraction=fraction))

        return cls._scale_unit_separator.join(parts)

    @classmethod
    def parse(cls, s: str, debug: bool = False) -> UnitBase:
        result = cls._do_parse(s, debug)
        if hasattr(result, "function_unit"):
            raise ValueError("Function units are not yet supported for FITS units.")
        return result



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Global Curvature.

This is private API. See `~astropy.cosmology.traits` for public API.

"""

__all__ = ["CurvatureComponent"]

import abc

import numpy as np
from numpy.typing import ArrayLike, NDArray

from astropy.cosmology._src.utils import aszarr, deprecated_keywords
from astropy.units import Quantity


class CurvatureComponent:
    """The object has attributes and methods related to the global curvature.

    This is a trait class; it is not meant to be instantiated directly, but
    instead to be used as a mixin to other classes.

    """

    @property
    @abc.abstractmethod
    def Ok0(self) -> float | np.floating:
        """Omega curvature; the effective curvature density/critical density at z=0."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def is_flat(self) -> bool:
        """Return `bool`; `True` if the cosmology is globally flat."""
        raise NotImplementedError

    @deprecated_keywords("z", since="7.0")
    def Ok(self, z: Quantity | ArrayLike) -> NDArray[np.floating]:
        """Return the equivalent density parameter for curvature at redshift ``z``.

        Parameters
        ----------
        z : Quantity-like ['redshift'], array-like
            Input redshift.

            .. versionchanged:: 7.0
                Passing z as a keyword argument is deprecated.

        Returns
        -------
        Ok : ndarray
            The equivalent density parameter for curvature at each redshift.

            .. versionchanged:: 7.2
                Always returns a numpy object, never a `float`.

        Examples
        --------
        >>> import numpy as np
        >>> from astropy.cosmology import Planck18, units as cu

        >>> Planck18.Ok(2)
        array(0.)

        >>> Planck18.Ok([1, 2])
        array([0., 0.])

        >>> Planck18.Ok(np.array([2]))
        array([0.])

        >>> Planck18.Ok(2 * cu.redshift)
        array(0.)

        >>> cosmo = Planck18.clone(Ode0=0.71, to_nonflat=True)

        >>> cosmo.Ok0
        np.float64(-0.021153694455455927)

        >>> cosmo.Ok(100)
        np.float64(-0.0006557825253017665)

        """
        z = aszarr(z)
        if self.Ok0 == 0:  # Common enough to be worth checking explicitly
            return np.zeros(getattr(z, "shape", ()))
        return self.Ok0 * (z + 1.0) ** 2 * self.inv_efunc(z) ** 2



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Miscellaneous utilities for `astropy.units`.

None of the functions in the module are meant for use outside of the
package.
"""

from fractions import Fraction
from typing import SupportsFloat

import numpy as np
from numpy import finfo

from .errors import UnitScaleError
from .typing import UnitPower, UnitPowerLike, UnitScale, UnitScaleLike

_float_finfo = finfo(float)
# take float here to ensure comparison with another float is fast
# give a little margin since often multiple calculations happened
_JUST_BELOW_UNITY = float(1.0 - 4.0 * _float_finfo.epsneg)
_JUST_ABOVE_UNITY = float(1.0 + 4.0 * _float_finfo.eps)


def is_effectively_unity(value: UnitScaleLike) -> bool | np.bool_:
    # value is *almost* always real, except, e.g., for u.mag**0.5, when
    # it will be complex.  Use try/except to ensure normal case is fast
    try:
        return _JUST_BELOW_UNITY <= value <= _JUST_ABOVE_UNITY
    except TypeError:  # value is complex
        return (
            _JUST_BELOW_UNITY <= value.real <= _JUST_ABOVE_UNITY
            and _JUST_BELOW_UNITY <= value.imag + 1 <= _JUST_ABOVE_UNITY
        )


def sanitize_scale(scale: UnitScaleLike) -> UnitScale:
    if is_effectively_unity(scale):
        return 1.0
    if not scale:
        raise UnitScaleError("cannot create a unit with a scale of 0.")
    if type(scale) is float:  # float is very common, so handle it fast
        return scale
    if isinstance(scale, SupportsFloat):
        return float(scale)

    if abs(scale.real) > abs(scale.imag):
        if is_effectively_unity(scale.imag / scale.real + 1):
            return float(scale.real)
    elif is_effectively_unity(scale.real / scale.imag + 1):
        return complex(0.0, scale.imag)
    return complex(scale)


def maybe_simple_fraction(p: UnitPowerLike, max_denominator: int = 100) -> UnitPower:
    """Fraction very close to x with denominator at most max_denominator.

    The fraction has to be such that fraction/x is unity to within 4 ulp.
    If such a fraction does not exist, returns the float number.

    The algorithm is that of `fractions.Fraction.limit_denominator`, but
    sped up by not creating a fraction to start with.

    If the input is zero, an integer or `fractions.Fraction`, just return it.
    """
    if p.__class__ is int or p.__class__ is Fraction:
        return p
    if p == 0:
        return 0  # p might be some numpy number, but we want a Python int
    n, d = float(p).as_integer_ratio()
    a = n // d
    # Normally, start with 0,1 and 1,0; here we have applied first iteration.
    n0, d0 = 1, 0
    n1, d1 = a, 1
    while d1 <= max_denominator:
        if _JUST_BELOW_UNITY <= n1 / (d1 * p) <= _JUST_ABOVE_UNITY:
            return Fraction(n1, d1)
        n, d = d, n - a * d
        a = n // d
        n0, n1 = n1, n0 + a * n1
        d0, d1 = d1, d0 + a * d1

    return float(p)


def sanitize_power(p: UnitPowerLike) -> UnitPower:
    """Convert the power to a float, an integer, or a Fraction.

    If a fractional power can be represented exactly as a floating point
    number, convert it to a float, to make the math much faster; otherwise,
    retain it as a `fractions.Fraction` object to avoid losing precision.
    Conversely, if the value is indistinguishable from a rational number with a
    low-numbered denominator, convert to a Fraction object.
    If a power can be represented as an integer, use that.

    Parameters
    ----------
    p : float, int, Rational, Fraction
        Power to be converted.
    """
    if p.__class__ is int:
        return p

    denom = getattr(p, "denominator", None)
    if denom is None:
        # This returns either a (simple) Fraction or the same float.
        p = maybe_simple_fraction(p)
        # If still a float, nothing more to be done.
        if isinstance(p, float):
            return p

        # Otherwise, check for simplifications.
        denom = p.denominator

    if denom == 1:
        return int(p.numerator)

    elif (denom & (denom - 1)) == 0:
        # Above is a bit-twiddling hack to see if denom is a power of two.
        # If so, float does not lose precision and will speed things up.
        p = float(p)

    return p


def resolve_fractions(
    a: UnitPowerLike, b: UnitPowerLike
) -> tuple[UnitPowerLike, UnitPowerLike]:
    """
    If either input is a Fraction, convert the other to a Fraction
    (at least if it does not have a ridiculous denominator).
    This ensures that any operation involving a Fraction will use
    rational arithmetic and preserve precision.
    """
    # We short-circuit on the most common cases of int and float, since
    # isinstance(a, Fraction) is very slow for any non-Fraction instances.
    a_is_fraction = (
        a.__class__ is not int and a.__class__ is not float and isinstance(a, Fraction)
    )
    b_is_fraction = (
        b.__class__ is not int and b.__class__ is not float and isinstance(b, Fraction)
    )
    if a_is_fraction and not b_is_fraction:
        b = maybe_simple_fraction(b)
    elif not a_is_fraction and b_is_fraction:
        a = maybe_simple_fraction(a)
    return a, b



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from copy import copy, deepcopy

import numpy as np
import pytest

from astropy import wcs

_WCS_UNDEFINED = 987654321.0e99


def test_celprm_init():
    # test PyCelprm_cnew
    assert wcs.WCS().wcs.cel

    # test PyCelprm_new
    assert wcs.Celprm()

    with pytest.raises(wcs.InvalidPrjParametersError):
        cel = wcs.Celprm()
        cel.set()

    # test deletion does not crash
    cel = wcs.Celprm()
    del cel


def test_celprm_copy():
    # shallow copy
    cel = wcs.Celprm()
    cel2 = copy(cel)
    cel3 = copy(cel2)
    cel.ref = [6, 8, 18, 3]
    assert np.allclose(cel.ref, cel2.ref, atol=1e-12, rtol=0) and np.allclose(
        cel.ref, cel3.ref, atol=1e-12, rtol=0
    )
    del cel, cel2, cel3

    # deep copy
    cel = wcs.Celprm()
    cel2 = deepcopy(cel)
    cel.ref = [6, 8, 18, 3]
    assert not np.allclose(cel.ref, cel2.ref, atol=1e-12, rtol=0)
    del cel, cel2


def test_celprm_offset():
    cel = wcs.Celprm()
    assert not cel.offset
    cel.offset = True
    assert cel.offset


def test_celprm_prj():
    cel = wcs.Celprm()
    assert cel.prj is not None
    cel.prj.code = "TAN"
    cel.set()
    assert cel._flag


def test_celprm_phi0():
    cel = wcs.Celprm()
    cel.prj.code = "TAN"

    assert cel.phi0 == None
    assert cel._flag == 0
    cel.set()
    assert cel.phi0 == 0.0

    cel.phi0 = 0.0
    assert cel._flag

    cel.phi0 = 2.0
    assert cel._flag == 0

    cel.phi0 = None
    assert cel.phi0 == None
    assert cel._flag == 0


def test_celprm_theta0():
    cel = wcs.Celprm()
    cel.prj.code = "TAN"

    assert cel.theta0 == None
    assert cel._flag == 0

    cel.theta0 = 4.0
    cel.set()
    assert cel.theta0 == 4.0

    cel.theta0 = 4.0
    assert cel._flag

    cel.theta0 = 8.0
    assert cel._flag == 0

    cel.theta0 = None
    assert cel.theta0 == None
    assert cel._flag == 0


def test_celprm_ref():
    cel = wcs.Celprm()
    cel.prj.code = "TAN"
    cel.set()

    assert np.allclose(cel.ref, [0.0, 0.0, 180.0, 0.0], atol=1e-12, rtol=0)

    cel.phi0 = 2.0
    cel.theta0 = 4.0
    cel.ref = [123, 12]
    cel.set()

    assert np.allclose(cel.ref, [123.0, 12.0, 2, 82], atol=1e-12, rtol=0)

    cel.ref = [None, 13, None, None]
    assert np.allclose(cel.ref, [123.0, 13.0, 2, 82], atol=1e-12, rtol=0)


def test_celprm_isolat():
    cel = wcs.Celprm()
    cel.prj.code = "TAN"
    cel.set()

    assert cel.isolat == 0


def test_celprm_latpreq():
    cel = wcs.Celprm()
    cel.prj.code = "TAN"
    cel.set()

    assert cel.latpreq == 0


def test_celprm_euler():
    cel = wcs.Celprm()
    cel.prj.code = "TAN"
    cel.set()

    assert np.allclose(cel.euler, [0.0, 90.0, 180.0, 0.0, 1.0], atol=1e-12, rtol=0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Contains the transformation functions for getting to "observed" systems from ICRS.
"""

import erfa

from astropy import units as u
from astropy.coordinates.baseframe import frame_transform_graph
from astropy.coordinates.builtin_frames.utils import atciqz, aticq
from astropy.coordinates.erfa_astrom import erfa_astrom
from astropy.coordinates.representation import (
    CartesianRepresentation,
    SphericalRepresentation,
    UnitSphericalRepresentation,
)
from astropy.coordinates.transformations import FunctionTransformWithFiniteDifference
from astropy.utils.compat import COPY_IF_NEEDED

from .altaz import AltAz
from .hadec import HADec
from .icrs import ICRS
from .utils import PIOVER2


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, ICRS, AltAz)
@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, ICRS, HADec)
def icrs_to_observed(icrs_coo, observed_frame):
    # if the data are UnitSphericalRepresentation, we can skip the distance calculations
    is_unitspherical = (
        isinstance(icrs_coo.data, UnitSphericalRepresentation)
        or icrs_coo.cartesian.x.unit == u.one
    )
    # first set up the astrometry context for ICRS<->observed
    astrom = erfa_astrom.get().apco(observed_frame)

    # correct for parallax to find BCRS direction from observer (as in erfa.pmpx)
    if is_unitspherical:
        srepr = icrs_coo.spherical
    else:
        observer_icrs = CartesianRepresentation(
            astrom["eb"], unit=u.au, xyz_axis=-1, copy=COPY_IF_NEEDED
        )
        srepr = (icrs_coo.cartesian - observer_icrs).represent_as(
            SphericalRepresentation
        )

    # convert to topocentric CIRS
    cirs_ra, cirs_dec = atciqz(srepr, astrom)

    # now perform observed conversion
    if isinstance(observed_frame, AltAz):
        lon, zen, _, _, _ = erfa.atioq(cirs_ra, cirs_dec, astrom)
        lat = PIOVER2 - zen
    else:
        _, _, lon, lat, _ = erfa.atioq(cirs_ra, cirs_dec, astrom)

    if is_unitspherical:
        obs_srepr = UnitSphericalRepresentation(
            lon << u.radian, lat << u.radian, copy=False
        )
    else:
        obs_srepr = SphericalRepresentation(
            lon << u.radian, lat << u.radian, srepr.distance, copy=False
        )
    return observed_frame.realize_frame(obs_srepr)


@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, AltAz, ICRS)
@frame_transform_graph.transform(FunctionTransformWithFiniteDifference, HADec, ICRS)
def observed_to_icrs(observed_coo, icrs_frame):
    # if the data are UnitSphericalRepresentation, we can skip the distance calculations
    is_unitspherical = (
        isinstance(observed_coo.data, UnitSphericalRepresentation)
        or observed_coo.cartesian.x.unit == u.one
    )

    usrepr = observed_coo.represent_as(UnitSphericalRepresentation)
    lon = usrepr.lon.to_value(u.radian)
    lat = usrepr.lat.to_value(u.radian)

    if isinstance(observed_coo, AltAz):
        # the 'A' indicates zen/az inputs
        coord_type = "A"
        lat = PIOVER2 - lat
    else:
        coord_type = "H"

    # first set up the astrometry context for ICRS<->CIRS at the observed_coo time
    astrom = erfa_astrom.get().apco(observed_coo)

    # Topocentric CIRS
    cirs_ra, cirs_dec = erfa.atoiq(coord_type, lon, lat, astrom) << u.radian
    if is_unitspherical:
        srepr = SphericalRepresentation(cirs_ra, cirs_dec, 1, copy=COPY_IF_NEEDED)
    else:
        srepr = SphericalRepresentation(
            lon=cirs_ra,
            lat=cirs_dec,
            distance=observed_coo.distance,
            copy=COPY_IF_NEEDED,
        )

    # BCRS (Astrometric) direction to source
    bcrs_ra, bcrs_dec = aticq(srepr, astrom) << u.radian

    # Correct for parallax to get ICRS representation
    if is_unitspherical:
        icrs_srepr = UnitSphericalRepresentation(bcrs_ra, bcrs_dec, copy=COPY_IF_NEEDED)
    else:
        icrs_srepr = SphericalRepresentation(
            lon=bcrs_ra,
            lat=bcrs_dec,
            distance=observed_coo.distance,
            copy=COPY_IF_NEEDED,
        )
        observer_icrs = CartesianRepresentation(
            astrom["eb"], unit=u.au, xyz_axis=-1, copy=COPY_IF_NEEDED
        )
        newrepr = icrs_srepr.to_cartesian() + observer_icrs
        icrs_srepr = newrepr.represent_as(SphericalRepresentation)

    return icrs_frame.realize_frame(icrs_srepr)


# Create loopback transformations
frame_transform_graph._add_merged_transform(AltAz, ICRS, AltAz)
frame_transform_graph._add_merged_transform(HADec, ICRS, HADec)
# for now we just implement this through ICRS to make sure we get everything
# covered
# Before, this was using CIRS as intermediate frame, however this is much
# slower than the direct observed<->ICRS transform added in 4.3
# due to how the frame attribute broadcasting works, see
# https://github.com/astropy/astropy/pull/10994#issuecomment-722617041



# Licensed under a 3-clause BSD style license - see LICENSE.rst

from itertools import combinations

import numpy as np
import pytest

from astropy import units as u
from astropy.coordinates import (
    CartesianDifferential,
    CartesianRepresentation,
    EarthLocation,
    SkyCoord,
    galactocentric_frame_defaults,
)
from astropy.coordinates.builtin_frames import (
    CIRS,
    FK4,
    FK5,
    GCRS,
    HCRS,
    ICRS,
    LSR,
    FK4NoETerms,
    Galactic,
    GalacticLSR,
    Galactocentric,
    Supergalactic,
)
from astropy.coordinates.distances import Distance
from astropy.tests.helper import assert_quantity_allclose as assert_allclose
from astropy.time import Time
from astropy.units import allclose

# used below in the next parametrized test
M31_DIST = Distance(770, u.kpc)
M31_COORDS = (
    ICRS(10.6847929 * u.deg, 41.2690650 * u.deg, M31_DIST),
    FK5(10.6847929 * u.deg, 41.2690650 * u.deg, M31_DIST),
    FK4(10.0004738 * u.deg, 40.9952444 * u.deg, M31_DIST),
    Galactic(121.1744050 * u.deg, -21.5729360 * u.deg, M31_DIST),
)


@pytest.mark.parametrize(
    "fromsys,tosys", tuple(combinations(M31_COORDS, 2)), ids=lambda x: type(x).__name__
)
def test_m31_coord_transforms(fromsys, tosys):
    """
    This tests a variety of coordinate conversions for the Chandra point-source
    catalog location of M31 from NED.
    """
    converted = fromsys.transform_to(tosys)
    expected = tosys.spherical
    assert_allclose(converted.spherical.lon, expected.lon, atol=0.1 * u.arcsec, rtol=0)
    assert_allclose(converted.spherical.lat, expected.lat, atol=0.25 * u.arcsec, rtol=0)
    assert_allclose(converted.distance, 770 * u.kpc)

    # check round-tripping
    roundtripped = converted.transform_to(fromsys).spherical
    expected = fromsys.spherical
    assert_allclose(roundtripped.lon, expected.lon, atol=0.1 * u.arcsec, rtol=0)
    assert_allclose(roundtripped.lat, expected.lat, atol=0.25 * u.arcsec, rtol=0)
    assert_allclose(roundtripped.distance, 770 * u.kpc)


def test_precession():
    """
    Ensures that FK4 and FK5 coordinates precess their equinoxes
    """
    j2000 = Time("J2000")
    b1950 = Time("B1950")
    j1975 = Time("J1975")
    b1975 = Time("B1975")

    fk4 = FK4(ra=1 * u.radian, dec=0.5 * u.radian)
    assert fk4.equinox.byear == b1950.byear
    fk4_2 = fk4.transform_to(FK4(equinox=b1975))
    assert fk4_2.equinox.byear == b1975.byear

    fk5 = FK5(ra=1 * u.radian, dec=0.5 * u.radian)
    assert fk5.equinox.jyear == j2000.jyear
    fk5_2 = fk5.transform_to(FK4(equinox=j1975))
    assert fk5_2.equinox.jyear == j1975.jyear


def test_fk5_galactic():
    """
    Check that FK5 -> Galactic gives the same as FK5 -> FK4 -> Galactic.
    """

    fk5 = FK5(ra=1 * u.deg, dec=2 * u.deg)

    direct = fk5.transform_to(Galactic())
    indirect = fk5.transform_to(FK4()).transform_to(Galactic())

    assert direct.separation(indirect).degree < 1.0e-10

    direct = fk5.transform_to(Galactic())
    indirect = fk5.transform_to(FK4NoETerms()).transform_to(Galactic())

    assert direct.separation(indirect).degree < 1.0e-10


def test_galactocentric():
    # when z_sun=0, transformation should be very similar to Galactic
    icrs_coord = ICRS(
        ra=np.linspace(0, 360, 10) * u.deg,
        dec=np.linspace(-90, 90, 10) * u.deg,
        distance=1.0 * u.kpc,
    )

    g_xyz = icrs_coord.transform_to(Galactic()).cartesian.xyz
    with galactocentric_frame_defaults.set("pre-v4.0"):
        gc_xyz = icrs_coord.transform_to(Galactocentric(z_sun=0 * u.kpc)).cartesian.xyz
    diff = np.abs(g_xyz - gc_xyz)

    assert allclose(diff[0], 8.3 * u.kpc, atol=1e-5 * u.kpc)
    assert allclose(diff[1:], 0 * u.kpc, atol=1e-5 * u.kpc)

    # generate some test coordinates
    g = Galactic(
        l=[0, 0, 45, 315] * u.deg,
        b=[-45, 45, 0, 0] * u.deg,
        distance=[np.sqrt(2)] * 4 * u.kpc,
    )
    with galactocentric_frame_defaults.set("pre-v4.0"):
        xyz = g.transform_to(
            Galactocentric(galcen_distance=1.0 * u.kpc, z_sun=0.0 * u.pc)
        ).cartesian.xyz
    true_xyz = np.array([[0, 0, -1.0], [0, 0, 1], [0, 1, 0], [0, -1, 0]]).T * u.kpc
    assert allclose(xyz.to(u.kpc), true_xyz.to(u.kpc), atol=1e-5 * u.kpc)

    # check that ND arrays work

    # from Galactocentric to Galactic
    x = np.linspace(-10.0, 10.0, 100) * u.kpc
    y = np.linspace(-10.0, 10.0, 100) * u.kpc
    z = np.zeros_like(x)

    # from Galactic to Galactocentric
    l = np.linspace(15, 30.0, 100) * u.deg
    b = np.linspace(-10.0, 10.0, 100) * u.deg
    d = np.ones_like(l.value) * u.kpc

    with galactocentric_frame_defaults.set("latest"):
        g1 = Galactocentric(x=x, y=y, z=z)
        g2 = Galactocentric(
            x=x.reshape(100, 1, 1), y=y.reshape(100, 1, 1), z=z.reshape(100, 1, 1)
        )

        g1t = g1.transform_to(Galactic())
        g2t = g2.transform_to(Galactic())

        assert_allclose(g1t.cartesian.xyz, g2t.cartesian.xyz[:, :, 0, 0])

        g1 = Galactic(l=l, b=b, distance=d)
        g2 = Galactic(
            l=l.reshape(100, 1, 1),
            b=b.reshape(100, 1, 1),
            distance=d.reshape(100, 1, 1),
        )

        g1t = g1.transform_to(Galactocentric())
        g2t = g2.transform_to(Galactocentric())

        np.testing.assert_almost_equal(
            g1t.cartesian.xyz.value, g2t.cartesian.xyz.value[:, :, 0, 0]
        )


def test_supergalactic():
    """
    Check Galactic<->Supergalactic and Galactic<->ICRS conversion.
    """
    # Check supergalactic North pole.
    npole = Galactic(l=47.37 * u.degree, b=+6.32 * u.degree)
    assert allclose(npole.transform_to(Supergalactic()).sgb.deg, +90, atol=1e-9)

    # Check the origin of supergalactic longitude.
    lon0 = Supergalactic(sgl=0 * u.degree, sgb=0 * u.degree)
    lon0_gal = lon0.transform_to(Galactic())
    assert allclose(lon0_gal.l.deg, 137.37, atol=1e-9)
    assert allclose(lon0_gal.b.deg, 0, atol=1e-9)

    # Test Galactic<->ICRS with some positions that appear in Foley et al. 2008
    # (https://ui.adsabs.harvard.edu/abs/2008A%26A...484..143F)

    # GRB 021219
    supergalactic = Supergalactic(sgl=29.91 * u.degree, sgb=+73.72 * u.degree)
    icrs = SkyCoord("18h50m27s +31d57m17s")
    assert supergalactic.separation(icrs) < 0.005 * u.degree

    # GRB 030320
    supergalactic = Supergalactic(sgl=-174.44 * u.degree, sgb=+46.17 * u.degree)
    icrs = SkyCoord("17h51m36s -25d18m52s")
    assert supergalactic.separation(icrs) < 0.005 * u.degree


class TestHCRS:
    """
    Check HCRS<->ICRS coordinate conversions.

    Uses ICRS Solar positions predicted by get_body_barycentric; with `t1` and
    `tarr` as defined below, the ICRS Solar positions were predicted using, e.g.
    coord.ICRS(coord.get_body_barycentric(tarr, 'sun')).
    """

    def setup_method(self):
        self.t1 = Time("2013-02-02T23:00")
        self.t2 = Time("2013-08-02T23:00")
        self.tarr = Time(["2013-02-02T23:00", "2013-08-02T23:00"])

        self.sun_icrs_scalar = ICRS(
            ra=244.52984668 * u.deg,
            dec=-22.36943723 * u.deg,
            distance=406615.66347377 * u.km,
        )
        # array of positions corresponds to times in `tarr`
        self.sun_icrs_arr = ICRS(
            ra=[244.52989062, 271.40976248] * u.deg,
            dec=[-22.36943605, -25.07431079] * u.deg,
            distance=[406615.66347377, 375484.13558956] * u.km,
        )

        # corresponding HCRS positions
        self.sun_hcrs_t1 = HCRS(
            CartesianRepresentation([0.0, 0.0, 0.0] * u.km), obstime=self.t1
        )
        twod_rep = CartesianRepresentation([[0.0, 0.0], [0.0, 0.0], [0.0, 0.0]] * u.km)
        self.sun_hcrs_tarr = HCRS(twod_rep, obstime=self.tarr)
        self.tolerance = 5 * u.km

    def test_from_hcrs(self):
        # test scalar transform
        transformed = self.sun_hcrs_t1.transform_to(ICRS())
        separation = transformed.separation_3d(self.sun_icrs_scalar)
        assert_allclose(separation, 0 * u.km, atol=self.tolerance)

        # test non-scalar positions and times
        transformed = self.sun_hcrs_tarr.transform_to(ICRS())
        separation = transformed.separation_3d(self.sun_icrs_arr)
        assert_allclose(separation, 0 * u.km, atol=self.tolerance)

    def test_from_icrs(self):
        # scalar positions
        transformed = self.sun_icrs_scalar.transform_to(HCRS(obstime=self.t1))
        separation = transformed.separation_3d(self.sun_hcrs_t1)
        assert_allclose(separation, 0 * u.km, atol=self.tolerance)
        # nonscalar positions
        transformed = self.sun_icrs_arr.transform_to(HCRS(obstime=self.tarr))
        separation = transformed.separation_3d(self.sun_hcrs_tarr)
        assert_allclose(separation, 0 * u.km, atol=self.tolerance)


class TestHelioBaryCentric:
    """
    Check GCRS<->Heliocentric and Barycentric coordinate conversions.

    Uses the WHT observing site (information grabbed from data/sites.json).
    """

    def setup_method(self):
        wht = EarthLocation(342.12 * u.deg, 28.758333333333333 * u.deg, 2327 * u.m)
        self.obstime = Time("2013-02-02T23:00")
        self.wht_itrs = wht.get_itrs(obstime=self.obstime)

    def test_heliocentric(self):
        gcrs = self.wht_itrs.transform_to(GCRS(obstime=self.obstime))
        helio = gcrs.transform_to(HCRS(obstime=self.obstime))
        # Check it doesn't change from previous times.
        previous = [-1.02597256e11, 9.71725820e10, 4.21268419e10] * u.m
        assert_allclose(helio.cartesian.xyz, previous)

        # And that it agrees with SLALIB to within 14km
        helio_slalib = [-0.685820296, 0.6495585893, 0.2816005464] * u.au
        assert np.sqrt(((helio.cartesian.xyz - helio_slalib) ** 2).sum()) < 14.0 * u.km

    def test_barycentric(self):
        gcrs = self.wht_itrs.transform_to(GCRS(obstime=self.obstime))
        bary = gcrs.transform_to(ICRS())
        previous = [-1.02758958e11, 9.68331109e10, 4.19720938e10] * u.m
        assert_allclose(bary.cartesian.xyz, previous)

        # And that it agrees with SLALIB answer to within 14km
        bary_slalib = [-0.6869012079, 0.6472893646, 0.2805661191] * u.au
        assert np.sqrt(((bary.cartesian.xyz - bary_slalib) ** 2).sum()) < 14.0 * u.km


def test_lsr_sanity():
    # random numbers, but zero velocity in ICRS frame
    icrs = ICRS(
        ra=15.1241 * u.deg,
        dec=17.5143 * u.deg,
        distance=150.12 * u.pc,
        pm_ra_cosdec=0 * u.mas / u.yr,
        pm_dec=0 * u.mas / u.yr,
        radial_velocity=0 * u.km / u.s,
    )
    lsr = icrs.transform_to(LSR())

    lsr_diff = lsr.data.differentials["s"]
    cart_lsr_vel = lsr_diff.represent_as(CartesianRepresentation, base=lsr.data)
    lsr_vel = ICRS(cart_lsr_vel)
    gal_lsr = lsr_vel.transform_to(Galactic()).cartesian.xyz
    assert allclose(gal_lsr.to(u.km / u.s, u.dimensionless_angles()), lsr.v_bary.d_xyz)

    # moving with LSR velocity
    lsr = LSR(
        ra=15.1241 * u.deg,
        dec=17.5143 * u.deg,
        distance=150.12 * u.pc,
        pm_ra_cosdec=0 * u.mas / u.yr,
        pm_dec=0 * u.mas / u.yr,
        radial_velocity=0 * u.km / u.s,
    )
    icrs = lsr.transform_to(ICRS())

    icrs_diff = icrs.data.differentials["s"]
    cart_vel = icrs_diff.represent_as(CartesianRepresentation, base=icrs.data)
    vel = ICRS(cart_vel)
    gal_icrs = vel.transform_to(Galactic()).cartesian.xyz
    assert allclose(
        gal_icrs.to(u.km / u.s, u.dimensionless_angles()), -lsr.v_bary.d_xyz
    )


def test_hcrs_icrs_differentials():
    # Regression to ensure that we can transform velocities from HCRS to LSR.
    # Numbers taken from the original issue, gh-6835.
    hcrs = HCRS(
        ra=8.67 * u.deg,
        dec=53.09 * u.deg,
        distance=117 * u.pc,
        pm_ra_cosdec=4.8 * u.mas / u.yr,
        pm_dec=-15.16 * u.mas / u.yr,
        radial_velocity=23.42 * u.km / u.s,
    )
    icrs = hcrs.transform_to(ICRS())

    # The position and velocity should not change much
    assert allclose(hcrs.cartesian.xyz, icrs.cartesian.xyz, rtol=1e-8)
    assert allclose(hcrs.velocity.d_xyz, icrs.velocity.d_xyz, rtol=1e-2)

    hcrs2 = icrs.transform_to(HCRS())

    # The values should round trip
    assert allclose(hcrs.cartesian.xyz, hcrs2.cartesian.xyz, rtol=1e-12)
    assert allclose(hcrs.velocity.d_xyz, hcrs2.velocity.d_xyz, rtol=1e-12)


def test_cirs_icrs():
    """
    Test CIRS<->ICRS transformations, including self transform
    """
    t = Time("J2010")
    MOONDIST = 385000 * u.km  # approximate moon semi-major orbit axis of moon
    MOONDIST_CART = CartesianRepresentation(
        3**-0.5 * MOONDIST, 3**-0.5 * MOONDIST, 3**-0.5 * MOONDIST
    )

    loc = EarthLocation(lat=0 * u.deg, lon=0 * u.deg)
    cirs_geo_frame = CIRS(obstime=t)
    cirs_topo_frame = CIRS(obstime=t, location=loc)

    moon_geo = cirs_geo_frame.realize_frame(MOONDIST_CART)
    moon_topo = moon_geo.transform_to(cirs_topo_frame)

    # now check that the distance change is similar to earth radius
    assert (
        1000 * u.km
        < np.abs(moon_topo.distance - moon_geo.distance).to(u.au)
        < 7000 * u.km
    )

    # now check that it round-trips
    moon2 = moon_topo.transform_to(moon_geo)
    assert_allclose(moon_geo.cartesian.xyz, moon2.cartesian.xyz)

    # now check ICRS transform gives a decent distance from Barycentre
    moon_icrs = moon_geo.transform_to(ICRS())
    assert_allclose(moon_icrs.distance - 1 * u.au, 0.0 * u.R_sun, atol=3 * u.R_sun)


@pytest.mark.parametrize("frame", [LSR, GalacticLSR])
def test_lsr_loopback(frame):
    xyz = CartesianRepresentation(1, 2, 3) * u.AU
    xyz = xyz.with_differentials(CartesianDifferential(4, 5, 6) * u.km / u.s)

    v_bary = CartesianDifferential(5, 10, 15) * u.km / u.s

    # Test that the loopback properly handles a change in v_bary
    from_coo = frame(xyz)  # default v_bary
    to_frame = frame(v_bary=v_bary)

    explicit_coo = from_coo.transform_to(ICRS()).transform_to(to_frame)
    implicit_coo = from_coo.transform_to(to_frame)

    # Confirm that the explicit transformation changes the velocity but not the position
    assert allclose(explicit_coo.cartesian.xyz, from_coo.cartesian.xyz, rtol=1e-10)
    assert not allclose(
        explicit_coo.velocity.d_xyz, from_coo.velocity.d_xyz, rtol=1e-10
    )

    # Confirm that the loopback matches the explicit transformation
    assert allclose(explicit_coo.cartesian.xyz, implicit_coo.cartesian.xyz, rtol=1e-10)
    assert allclose(
        explicit_coo.velocity.d_xyz, implicit_coo.velocity.d_xyz, rtol=1e-10
    )


@pytest.mark.parametrize(
    "to_frame",
    [
        Galactocentric(galcen_coord=ICRS(300 * u.deg, -30 * u.deg)),
        Galactocentric(galcen_distance=10 * u.kpc),
        Galactocentric(z_sun=10 * u.pc),
        Galactocentric(roll=1 * u.deg),
    ],
)
def test_galactocentric_loopback(to_frame):
    xyz = CartesianRepresentation(1, 2, 3) * u.pc

    from_coo = Galactocentric(xyz)

    explicit_coo = from_coo.transform_to(ICRS()).transform_to(to_frame)
    implicit_coo = from_coo.transform_to(to_frame)

    # Confirm that the explicit transformation changes the position
    assert not allclose(explicit_coo.cartesian.xyz, from_coo.cartesian.xyz, rtol=1e-10)

    # Confirm that the loopback matches the explicit transformation
    assert allclose(explicit_coo.cartesian.xyz, implicit_coo.cartesian.xyz, rtol=1e-10)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import erfa
import numpy as np

from astropy import units as u
from astropy.coordinates.angles import Latitude, Longitude
from astropy.utils.decorators import format_doc

from .base import BaseRepresentation
from .cartesian import CartesianRepresentation

ELLIPSOIDS = {}
"""Available ellipsoids (defined in erfam.h, with numbers exposed in erfa)."""
# Note: they get filled by the creation of the geodetic classes.


geodetic_base_doc = """{__doc__}

    Parameters
    ----------
    lon, lat : angle-like
        The longitude and latitude of the point(s), in angular units. The
        latitude should be between -90 and 90 degrees, and the longitude will
        be wrapped to an angle between 0 and 360 degrees. These can also be
        instances of `~astropy.coordinates.Angle` and either
        `~astropy.coordinates.Longitude` not `~astropy.coordinates.Latitude`,
        depending on the parameter.

    height : `~astropy.units.Quantity` ['length']
        The height to the point(s).

    copy : bool, optional
        If `True` (default), arrays will be copied. If `False`, arrays will
        be references, though possibly broadcast to ensure matching shapes.
"""


@format_doc(geodetic_base_doc)
class BaseGeodeticRepresentation(BaseRepresentation):
    """
    Base class for geodetic representations.

    Subclasses need to set attributes ``_equatorial_radius`` and ``_flattening``
    to quantities holding correct values (with units of length and dimensionless,
    respectively), or alternatively an ``_ellipsoid`` attribute to the relevant ERFA
    index (as passed in to `erfa.eform`). The geodetic latitude is defined by the
    angle between the vertical to the surface at a specific point of the spheroid and
    its projection onto the equatorial plane.
    """

    attr_classes = {"lon": Longitude, "lat": Latitude, "height": u.Quantity}

    def __init_subclass__(cls, **kwargs):
        if "_ellipsoid" in cls.__dict__:
            equatorial_radius, flattening = erfa.eform(getattr(erfa, cls._ellipsoid))
            cls._equatorial_radius = equatorial_radius * u.m
            cls._flattening = flattening * u.dimensionless_unscaled
            ELLIPSOIDS[cls._ellipsoid] = cls
        elif (
            "_equatorial_radius" not in cls.__dict__
            or "_flattening" not in cls.__dict__
        ):
            raise AttributeError(
                f"{cls.__name__} requires '_ellipsoid' or '_equatorial_radius' and '_flattening'."
            )
        super().__init_subclass__(**kwargs)

    def __init__(self, lon, lat=None, height=None, copy=True):
        if height is None and not isinstance(lon, self.__class__):
            height = 0 << u.m

        super().__init__(lon, lat, height, copy=copy)
        if not self.height.unit.is_equivalent(u.m):
            raise u.UnitTypeError(
                f"{self.__class__.__name__} requires height with units of length."
            )

    def to_cartesian(self):
        """
        Converts geodetic coordinates to 3D rectangular (geocentric)
        cartesian coordinates.
        """
        xyz = erfa.gd2gce(
            self._equatorial_radius,
            self._flattening,
            self.lon,
            self.lat,
            self.height,
        )
        return CartesianRepresentation(xyz, xyz_axis=-1, copy=False)

    @classmethod
    def from_cartesian(cls, cart):
        """
        Converts 3D rectangular cartesian coordinates (assumed geocentric) to
        geodetic coordinates.
        """
        # Compute geodetic/planetodetic angles
        lon, lat, height = erfa.gc2gde(
            cls._equatorial_radius, cls._flattening, cart.get_xyz(xyz_axis=-1)
        )
        return cls(lon, lat, height, copy=False)


@format_doc(geodetic_base_doc)
class BaseBodycentricRepresentation(BaseRepresentation):
    """Representation of points in bodycentric 3D coordinates.

    Subclasses need to set attributes ``_equatorial_radius`` and ``_flattening``
    to quantities holding correct values (with units of length and dimensionless,
    respectively). the bodycentric latitude and longitude are spherical latitude
    and longitude relative to the barycenter of the body.
    """

    attr_classes = {"lon": Longitude, "lat": Latitude, "height": u.Quantity}

    def __init_subclass__(cls, **kwargs):
        if (
            "_equatorial_radius" not in cls.__dict__
            or "_flattening" not in cls.__dict__
        ):
            raise AttributeError(
                f"{cls.__name__} requires '_equatorial_radius' and '_flattening'."
            )
        super().__init_subclass__(**kwargs)

    def __init__(self, lon, lat=None, height=None, copy=True):
        if height is None and not isinstance(lon, self.__class__):
            height = 0 << u.m

        super().__init__(lon, lat, height, copy=copy)
        if not self.height.unit.is_equivalent(u.m):
            raise u.UnitTypeError(
                f"{self.__class__.__name__} requires height with units of length."
            )

    def to_cartesian(self):
        """
        Converts bodycentric coordinates to 3D rectangular (geocentric)
        cartesian coordinates.
        """
        coslat = np.cos(self.lat)
        sinlat = np.sin(self.lat)
        coslon = np.cos(self.lon)
        sinlon = np.sin(self.lon)
        r = (
            self._equatorial_radius * np.hypot(coslat, (1 - self._flattening) * sinlat)
            + self.height
        )
        x = r * coslon * coslat
        y = r * sinlon * coslat
        z = r * sinlat
        return CartesianRepresentation(x=x, y=y, z=z, copy=False)

    @classmethod
    def from_cartesian(cls, cart):
        """
        Converts 3D rectangular cartesian coordinates (assumed geocentric) to
        bodycentric coordinates.
        """
        # Compute bodycentric latitude
        p = np.hypot(cart.x, cart.y)
        d = np.hypot(p, cart.z)
        lat = np.arctan2(cart.z, p)
        p_spheroid = cls._equatorial_radius * np.cos(lat)
        z_spheroid = cls._equatorial_radius * (1 - cls._flattening) * np.sin(lat)
        r_spheroid = np.hypot(p_spheroid, z_spheroid)
        height = d - r_spheroid
        lon = np.arctan2(cart.y, cart.x)
        return cls(lon, lat, height, copy=False)


@format_doc(geodetic_base_doc)
class WGS84GeodeticRepresentation(BaseGeodeticRepresentation):
    """Representation of points in WGS84 3D geodetic coordinates."""

    _ellipsoid = "WGS84"


@format_doc(geodetic_base_doc)
class WGS72GeodeticRepresentation(BaseGeodeticRepresentation):
    """Representation of points in WGS72 3D geodetic coordinates."""

    _ellipsoid = "WGS72"


@format_doc(geodetic_base_doc)
class GRS80GeodeticRepresentation(BaseGeodeticRepresentation):
    """Representation of points in GRS80 3D geodetic coordinates."""

    _ellipsoid = "GRS80"



import contextlib
import datetime
import decimal
import io
import sys
import textwrap
from pathlib import Path

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from astropy.io.misc.pyarrow.csv import strip_comment_lines
from astropy.table import Table
from astropy.utils.compat.optional_deps import HAS_BZ2, HAS_LZMA, HAS_PYARROW

if HAS_PYARROW:
    import pyarrow as pa
else:
    pytest.skip("pyarrow is not available", allow_module_level=True)


def check_tables_equal(t1, t2):
    """Check that two tables are equal.

    This function checks that the two tables have the same column names,
    lengths, and that the data in each column is equal. It also checks
    that the masks (if any) are equal.
    """
    assert t1.colnames == t2.colnames
    assert len(t1) == len(t2)
    for col1, col2 in zip(t1.itercols(), t2.itercols()):
        assert col1.name == col2.name
        assert col1.dtype == col2.dtype
        assert_array_equal(col1, col2)
        has_mask1 = hasattr(col1, "mask")
        has_mask2 = hasattr(col2, "mask")
        assert has_mask1 == has_mask2
        if has_mask1:
            assert_array_equal(col1.mask, col2.mask)


def convert_table_to_text(tbl, delimiter=",", **kwargs):
    """Convert a table to text using the io.ascii CSV format.

    This function is used to convert a table to text for testing purposes.
    It uses the `ascii.csv` format to write the table to a string buffer.
    """
    text_io = io.StringIO()
    tbl.write(text_io, format="ascii.csv", delimiter=delimiter, **kwargs)
    return text_io.getvalue()


@pytest.fixture(scope="module")
def tbl():
    """Masked table with bool, int, float, and non-ASCII string types."""
    return Table(
        rows=[
            [np.ma.masked, 1, np.ma.masked, np.ma.masked],
            [False, np.ma.masked, 2.5, 'bår q"ux'],
            [True, 2, 3.5, "bazœo"],
        ],
        names=["a", "b", "c 2", "d"],
    )


@pytest.fixture(scope="module")
def tbl_text(tbl):
    """Default CSV text representation for simple table above.

    This is equivalent to::

    '''
    a,b,ç 2,d
    ,,,
    False,1,2.5,"bår q""ux"
    True,2,3.5,bazœo
    '''
    """
    return convert_table_to_text(tbl)


@contextlib.contextmanager
def get_input_file(
    text: str,
    input_type: str,
    encoding: str | None = None,
    tmp_path: Path | None = None,
):
    """
    Generate an input file or stream based on the specified type and encoding.

    Parameters
    ----------
    text : str
        The text content to be written to the input file or stream.
    input_type : str
        The type of input to generate. Supported values are: - "str": A temporary file
        path as a string. - "path": A temporary file path as a `Path` object. -
        "bytesio": A `BytesIO` stream.
    encoding : str | None, optional
        The encoding to use when writing the text. If None, no encoding is applied.
    tmp_path : Path | None, optional
        The temporary path to use for the input file for "str" or "path" input type.

    Yields
    ------
    str | Path | io.BytesIO
        The generated input file or stream based on the specified `input_type`.

    Notes
    -----
    - For "str" and "path" input types, a temporary file in ``tmp_path`` is created and
      its content is written with the specified encoding (if provided).
    - For "bytesio", the text is encoded into bytes and returned as a `BytesIO` stream.
    - The temporary file is automatically cleaned up after use.
    """
    encoding = encoding or "utf-8"
    if input_type in ("str", "path"):
        path = tmp_path / "test.csv"
        path.write_text(text, encoding=encoding)
        yield str(path) if input_type == "str" else path
    elif input_type == "bytesio":
        yield io.BytesIO(text.encode(encoding))
    else:
        raise ValueError(f"Unknown input_type: {input_type}")


def table_read_csv(
    text: str,
    input_type: str = "bytesio",
    encoding: str | None = None,
    tmp_path: Path | None = None,
    **kwargs,
):
    """Read ``text`` using ``Table.read`` with format="pyarrow.csv".

    The ``input_type`` parameter determines how the text is passed to ``Table.read``.
    For "str" or "path", a named temporary file is created and that name ("str") or path
    ("path") is passed to ``Table.read``. For "bytesio", a ``BytesIO`` object is created
    and passed to ``Table.read``. The text is encoded using the specified encoding (if
    any) before being passed to ``BytesIO``. The default is "bytesio".

    The ``encoding`` parameter is used only if ``input_type`` is "bytesio". If
    ``encoding`` is None, the text is encoded using the default encoding.

    Parameters
    ----------
    text : str
        The text to read.
    input_type : str
        The type of input for Table.read(). One of "str", "path", or "bytesio".
    encoding : str | None
        The encoding to use when reading the text. If None, no encoding is applied.
    tmp_path : Path | None
        The temporary path to use for the input file for input_type="str" or "path".
    **kwargs : dict
        Additional keyword arguments to pass to Table.read().
    """
    if encoding is not None:
        kwargs["encoding"] = encoding

    with get_input_file(text, input_type, encoding, tmp_path) as input_file:
        return Table.read(input_file, format="pyarrow.csv", **kwargs)


@pytest.mark.parametrize("input_type", ["str", "path", "bytesio"])
@pytest.mark.parametrize("encoding", [None, "utf-8", "utf-16"])
def test_read_tbl_simple_input_type_encoding(
    input_type, encoding, tbl, tbl_text, tmp_path
):
    """Test reading a simple CSV file with different input types.

    This tests:
    - input_file : str, PathLike, or binary file-like object
    - encoding : None, "utf-8", or "utf-16"
    """
    out = table_read_csv(tbl_text, input_type, encoding, tmp_path=tmp_path)
    check_tables_equal(tbl, out)


@pytest.mark.parametrize("delimiter", ["|", "\t", " "])
def test_read_delimiter(tbl, delimiter):
    """Test reading a simple CSV file with different delimiters.

    This tests:
    - delimiter : "|", "\t", and " "
    """
    tbl_text = convert_table_to_text(tbl, delimiter=delimiter)
    out = table_read_csv(tbl_text, delimiter=delimiter)
    check_tables_equal(tbl, out)


def test_read_dtypes(tbl_text):
    """Test reading a simple CSV file with different input types.

    This tests:
    - dtypes: dict of types, with a sampling of types including numpy and pyarrow
    """
    tbl_text = textwrap.dedent(
        """
        a,b,c,d,e,f,time
        0,1,2,3,0,5,12:34:56.123456
        ,,,,,,
        """
    )

    dtypes = {
        "a": "str",
        "b": pa.decimal128(precision=5, scale=2),
        "c": np.uint8,
        "d": pa.float32(),
        "e": "bool",
        "f": pa.utf8(),
        "time": pa.time64("us"),
    }
    out = table_read_csv(tbl_text, dtypes=dtypes)
    assert out["a"].dtype == "U1"
    assert out["b"].dtype == "object"
    assert isinstance(out["b"][0], decimal.Decimal)
    assert out["c"].dtype == "uint8"
    assert out["d"].dtype == "float32"
    assert out["e"].dtype == "bool"
    assert out["f"].dtype == "U1"
    assert out["time"].dtype == "object"
    assert isinstance(out["time"][0], datetime.time)

    # Unmask the table and check the zero values
    for col in out.itercols():
        col.mask[:] = False
    assert out["a"][1] == ""
    assert out["b"][1] == decimal.Decimal(0)
    assert out["c"][1] == 0
    assert out["d"][1] == 0.0
    assert out["e"][1] == False  # noqa: E712
    assert out["f"][1] == ""
    assert out["time"][1] == datetime.time(0, 0, 0)


def test_read_dtypes_invalid_conversion(tbl_text):
    """Test reading with an dtype that is inconsistent with the file data.

    This tests:
    - dtypes: dict of types, with invalid type for conversion provided
    """
    dtypes = {"a": "int"}  # True/False are not convertible to int
    with pytest.raises(pa.lib.ArrowInvalid, match="In CSV column #0"):
        table_read_csv(tbl_text, dtypes=dtypes)


def test_read_dtypes_invalid_dtype(tbl_text):
    """Test reading with an invalid dtype.

    This tests:
    - dtypes: dict of types, with invalid type for conversion provided
    """
    dtypes = {"a": "asdf"}
    with pytest.raises(TypeError, match="data type 'asdf' not understood"):
        table_read_csv(tbl_text, dtypes=dtypes)


def test_read_quotechar():
    """Test reading a simple CSV file with a single quote quotechar.

    This tests:
    - quotechar : single quote and False
    """
    tbl_text = textwrap.dedent("""
    a,b,c
    0,'''',2.5
    ","2",'3.5'
    """)

    out = table_read_csv(tbl_text, quotechar="'")
    exp = [
        " a    b      c   ",
        "str1 str3 float64",
        "---- ---- -------",
        "   0    '     2.5",  # '''' turns into a single quote
        '   "  "2"     3.5',  # '3.5' is still float
    ]
    assert out.pformat(show_dtype=True) == exp

    out = table_read_csv(tbl_text, quotechar=False)
    # No quoting conversions with quotechar=False
    exp = [
        " a    b     c  ",
        "str1 str4  str5",
        "---- ---- -----",
        "   0 ''''   2.5",
        '   "  "2" \'3.5\'',
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_doublequote():
    """Test reading a simple CSV file with doublequote.

    This tests:
    - doublequote : False (default True is tested above)
    """
    tbl_text = textwrap.dedent('''
    a,b,c
    0,"""",2.5
    ''')
    out = table_read_csv(tbl_text, doublequote=False)
    exp = [
        "  a    b      c   ",
        "int64 str2 float64",
        "----- ---- -------",
        '    0   ""     2.5',
    ]
    assert out.pformat(show_dtype=True) == exp


@pytest.mark.parametrize("ec", ["\\", "/"])
def test_read_escapechar(ec):
    r"""Test reading a simple CSV file with escapechar.

    This tests:
    - escapechar : "\" and "/"
    - Blank lines are ignored (default behavior, tested elsewhere as well)
    """
    tbl_text = textwrap.dedent(f"""
    a,b,c
    0,{ec},,{ec}"{ec}'
    """)
    out = table_read_csv(tbl_text, escapechar=ec)
    exp = [
        "  a    b    c  ",
        "int64 str1 str2",
        "----- ---- ----",
        "    0    ,   \"'",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_header_start_1():
    """Test reading a simple CSV file with header_start.

    This tests:
    - header_start : 1
    """
    tbl_text = "a,b,c\n0,1,2\n3,4,5"
    out = table_read_csv(tbl_text, header_start=1)
    exp = [
        "  0     1     2  ",
        "int64 int64 int64",
        "----- ----- -----",
        "    3     4     5",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_header_start_none():
    """Test reading a simple CSV file with header_start.

    This tests:
    - header_start : None
    """
    tbl_text = "a,b,c\n0,1,2\n3,4,5"
    out = table_read_csv(tbl_text, header_start=None)
    exp = [
        " f0   f1   f2 ",
        "str1 str1 str1",
        "---- ---- ----",
        "   a    b    c",
        "   0    1    2",
        "   3    4    5",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_header_start_none_data_start_1():
    """Test reading a simple CSV file with data_start.

    This tests:
    - header_start : None
    - data_start : 1
    - names : provided list
    """
    tbl_text = "a,b,c\n0,1,2\n3,4,5\n6,7,8"
    names = ["x", "y", "z"]
    out = table_read_csv(tbl_text, header_start=None, data_start=1, names=names)
    exp = [
        "  x     y     z  ",
        "int64 int64 int64",
        "----- ----- -----",
        "    0     1     2",
        "    3     4     5",
        "    6     7     8",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_data_start_2():
    """Test reading a simple CSV file with data_start.

    This tests:
    - data_start : 2
    """
    tbl_text = "a,b,c\n0,1,2\n3,4,5\n6,7,8"
    out = table_read_csv(tbl_text, data_start=2)
    exp = [
        "  a     b     c  ",
        "int64 int64 int64",
        "----- ----- -----",
        "    3     4     5",
        "    6     7     8",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_names_exception(tbl_text):
    """Test that providing names without header_start=None raises an exception.

    This tests:
    - names : provided list
    """
    with pytest.raises(
        ValueError, match="cannot specify `names` unless `header_start=None`"
    ):
        table_read_csv(tbl_text, names=["x", "y"])


def test_read_include_names():
    tbl_text = "a,b,c\n0,1,2\n3,4,5"
    out = table_read_csv(tbl_text, include_names=["a", "b"])
    exp = [
        "  a     b  ",
        "int64 int64",
        "----- -----",
        "    0     1",
        "    3     4",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_include_names_invalid(tbl_text):
    with pytest.raises(
        pa.lib.ArrowKeyError,
        match="Column 'x' in include_columns does not exist in CSV file",
    ):
        table_read_csv(tbl_text, include_names=["x", "y"])


@pytest.mark.parametrize("on_top", [True, False])
def test_strip_comment_lines(on_top):
    """Low level test of stripping comment lines."""
    tbl_text = textwrap.dedent("""\
    # Comment 1
     # Comment 2 (leading whitespace)
    # Comment 3
    a,b,c
    0,1,2
    3,4,5
    """)
    if not on_top:
        tbl_text += "# Comment 4\n# Comment 5\n"

    exp = textwrap.dedent("""\
    a,b,c
    0,1,2
    3,4,5
    """).encode("utf-8")

    input_file = io.BytesIO(tbl_text.encode("utf-8"))

    # For the default header_start and data_start values, the behavior depends on
    # whether there comment lines are only at the top of the file or not.
    output_file, header_start = strip_comment_lines(
        input_file, comment="#", header_start=0, data_start=None, encoding="utf-8"
    )
    if on_top:
        assert output_file is None
        assert header_start == 3
    else:
        assert output_file.getvalue() == exp
        assert header_start is None

    # In both of the following cases, strip_comment_lines should return a new BytesIO
    # buffer with the comment lines stripped, and the header_start should be None.
    input_file.seek(0)
    output_file, header_start = strip_comment_lines(
        input_file, comment="#", header_start=1, data_start=2, encoding="utf-8"
    )
    assert output_file.getvalue() == exp
    assert header_start is None

    input_file.seek(0)
    output_file, header_start = strip_comment_lines(
        input_file, comment="#", header_start=None, data_start=1, encoding="utf-8"
    )
    assert output_file.getvalue() == exp
    assert header_start is None


@pytest.mark.parametrize("on_top", [True, False])
def test_read_comments(on_top):
    """Test reading a simple CSV file with comments.

    This tests:
    - comment : "#"
    - All comment lines are at the top of file
    """
    tbl_text = textwrap.dedent("""\
    # Comment 1
     # Comment 2 (leading whitespace)
    # Comment 3
    a,b,c
    0,1,2
    3,4,5
    """)
    if not on_top:
        tbl_text += "\n# Comment 4\n# Comment 5\n"

    out = table_read_csv(tbl_text, comment="#")
    exp = [
        "  a     b     c  ",
        "int64 int64 int64",
        "----- ----- -----",
        "    0     1     2",
        "    3     4     5",
    ]
    assert out.pformat(show_dtype=True) == exp

    out = table_read_csv(tbl_text, header_start=None, data_start=1, comment="#")
    exp = [
        "  f0    f1    f2 ",
        "int64 int64 int64",
        "----- ----- -----",
        "    0     1     2",
        "    3     4     5",
    ]
    assert out.pformat(show_dtype=True) == exp

    out = table_read_csv(tbl_text, header_start=1, data_start=2, comment="#")
    exp = [
        "  0     1     2  ",
        "int64 int64 int64",
        "----- ----- -----",
        "    3     4     5",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_null_values():
    """Test reading a simple CSV file with null values.

    This tests:
    - null_values : ["", "NLL"], ["NLL"], []
    """
    tbl_text = textwrap.dedent("""\
    a,b,c
    NLL,2,3
    4,,6
    """)
    out = table_read_csv(tbl_text, null_values=["", "NLL"])
    exp = Table(
        rows=[
            [np.ma.masked, 2, 3],
            [4, np.ma.masked, 6],
        ],
        names=["a", "b", "c"],
    )
    check_tables_equal(exp, out)

    out = table_read_csv(tbl_text, null_values=["NLL"])
    exp = Table(
        rows=[
            [np.ma.masked, "2", 3],
            [4, "", 6],
        ],
        names=["a", "b", "c"],
    )
    check_tables_equal(exp, out)

    out = table_read_csv(tbl_text, null_values=[])
    exp = Table(
        rows=[
            ["NLL", "2", 3],
            ["4", "", 6],
        ],
        names=["a", "b", "c"],
    )
    check_tables_equal(exp, out)


def test_read_null_values_default():
    """Test that the default null value is [""].

    PyArrow by default includes other null values like "NaN" or "nan".
    """
    tbl_text = textwrap.dedent("""\
    a,b,c
    ,nan,NaN
    4,,#N/A
    """)
    out = table_read_csv(tbl_text)
    exp = [
        "  a      b     c  ",
        "int64 float64 str4",
        "----- ------- ----",
        "   --     nan  NaN",
        "    4      -- #N/A",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_newlines_in_values():
    """Test reading a simple CSV file with newlines in values.

    This tests:
    - newlines_in_values : True (default is False)
    """
    tbl_text = textwrap.dedent("""\
    a,b,c
    0,"1
    2",3
    4,5,"6
    7"
    """)
    out = table_read_csv(tbl_text, newlines_in_values=True)
    exp = [
        "  a    b    c  ",
        "int64 str3 str3",
        "----- ---- ----",
        "    0 1\\n2    3",
        "    4    5 6\\n7",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_dates_times():
    """Test reading CSV file with times, dates, and timestamps."""
    tbl_text = textwrap.dedent("""
    date,time_of_day,timestamp_s,timestamp_ns
    2023-12-25,12:34:56,2023-12-25T12:34:56,2023-12-25T12:34:56.123456
    2024-01-01,23:59:59,2024-01-01T23:59:59,2024-01-01T23:59:59.987654
    """)
    out = table_read_csv(tbl_text)
    exp = [
        "     date     time_of_day     timestamp_s              timestamp_ns        ",
        "datetime64[D]    object      datetime64[s]            datetime64[ns]       ",
        "------------- ----------- ------------------- -----------------------------",
        "   2023-12-25    12:34:56 2023-12-25T12:34:56 2023-12-25T12:34:56.123456000",
        "   2024-01-01    23:59:59 2024-01-01T23:59:59 2024-01-01T23:59:59.987654000",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_dates_times_masked():
    """Test reading CSV file with missing times, dates, and timestamps."""
    tbl_text = textwrap.dedent("""
    date,time_of_day,timestamp_s,timestamp_ns
    ,,,
    2024-01-01,23:59:59,2024-01-01T23:59:59,2024-01-01T23:59:59.987654
    """)
    out = table_read_csv(tbl_text)
    exp = [
        "     date     time_of_day     timestamp_s              timestamp_ns        ",
        "datetime64[D]    object      datetime64[s]            datetime64[ns]       ",
        "------------- ----------- ------------------- -----------------------------",
        "           --          --                  --                            --",
        "   2024-01-01    23:59:59 2024-01-01T23:59:59 2024-01-01T23:59:59.987654000",
    ]
    assert out.pformat(show_dtype=True) == exp

    # Confirm the values under the mask
    for col in out.itercols():
        col.mask = False
    exp = [
        "     date     time_of_day     timestamp_s              timestamp_ns        ",
        "datetime64[D]    object      datetime64[s]            datetime64[ns]       ",
        "------------- ----------- ------------------- -----------------------------",
        "   1970-01-01    00:00:00 1970-01-01T00:00:00 1970-01-01T00:00:00.000000000",
        "   2024-01-01    23:59:59 2024-01-01T23:59:59 2024-01-01T23:59:59.987654000",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_dates_times_custom():
    """Test reading CSV file with custom timestamp formats."""
    tbl_text = textwrap.dedent("""
    date,timestamp_s,timestamp_custom
    2023-12-25,2023-12-25T12:34:56,01/31/2024 14:23:55
    2024-01-01,2024-01-01T23:59:59,12/25/2023 09:00:00
    """)
    out = table_read_csv(
        tbl_text, timestamp_parsers=[pa.csv.ISO8601, "%m/%d/%Y %H:%M:%S"]
    )
    exp = [
        "     date         timestamp_s       timestamp_custom ",
        "datetime64[D]    datetime64[s]       datetime64[s]   ",
        "------------- ------------------- -------------------",
        "   2023-12-25 2023-12-25T12:34:56 2024-01-31T14:23:55",
        "   2024-01-01 2024-01-01T23:59:59 2023-12-25T09:00:00",
    ]
    assert out.pformat(show_dtype=True) == exp


def test_read_whitespace_handling():
    """Test reading CSV file with leading and trailing whitespace.

    All whitespace in header and string columns are significant.
    Whitespace in a numeric column is ignored.
    """
    tbl_text = """
   a  , b , c
    0, 1.0 , x y
    3, 4.0 , w z """
    out = table_read_csv(tbl_text)
    assert out.colnames == ["   a  ", " b ", " c"]
    exp = [
        "   a      b       c ",
        "int64  float64  str5",
        "------ ------- -----",
        "     0     1.0   x y",
        "     3     4.0  w z ",
    ]
    assert out.pformat(show_dtype=True) == exp


@pytest.mark.xfail(sys.platform.startswith("win"), reason="Broken on Windows")
@pytest.mark.parametrize(
    "compressed_filename",
    [
        "test.csv.gz",
        pytest.param(
            "test.csv.bz2",
            marks=pytest.mark.xfail(not HAS_BZ2, reason="no bz2 support"),
        ),
        pytest.param(
            "test.csv.xz",
            marks=pytest.mark.xfail(not HAS_LZMA, reason="no lzma support"),
        ),
    ],
)
def test_compressed_files(tbl, tmp_path, compressed_filename):
    filename = tmp_path / "test.csv"
    tbl.write(filename)

    compressed_filename = tmp_path / compressed_filename
    if compressed_filename.suffix == ".gz":
        import gzip

        opener = gzip.open
    elif compressed_filename.suffix == ".bz2":
        import bz2

        opener = bz2.open
    elif compressed_filename.suffix == ".xz":
        import lzma

        opener = lzma.open
    else:
        # Shouldn't really happen
        opener = open

    # Compress ecsv file
    with open(filename, "rb") as f_in:
        with opener(compressed_filename, "wb") as f_out:
            f_out.writelines(f_in)

    # Open compressed file and compare to ensure it's read correctly
    t_comp = Table.read(compressed_filename, format="pyarrow.csv")
    assert_array_equal(tbl, t_comp)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
``fitsinfo`` is a command-line script based on astropy.io.fits for
printing a summary of the HDUs in one or more FITS files(s) to the
standard output.

Example usage of ``fitsinfo``:

1. Print a summary of the HDUs in a FITS file::

    $ fitsinfo filename.fits

    Filename: filename.fits
    No.    Name         Type      Cards   Dimensions   Format
    0    PRIMARY     PrimaryHDU     138   ()
    1    SCI         ImageHDU        61   (800, 800)   int16
    2    SCI         ImageHDU        61   (800, 800)   int16
    3    SCI         ImageHDU        61   (800, 800)   int16
    4    SCI         ImageHDU        61   (800, 800)   int16

2. Print a summary of HDUs of all the FITS files in the current directory::

    $ fitsinfo *.fits
"""

import argparse

import astropy.io.fits as fits
from astropy import __version__, log

DESCRIPTION = """
Print a summary of the HDUs in a FITS file(s).

This script is part of the Astropy package. See
https://docs.astropy.org/en/latest/io/fits/usage/scripts.html#module-astropy.io.fits.scripts.fitsinfo
for further documentation.
""".strip()


def fitsinfo(filename):
    """
    Print a summary of the HDUs in a FITS file.

    Parameters
    ----------
    filename : str
        The path to a FITS file.
    """
    try:
        fits.info(filename)
    except OSError as e:
        log.error(str(e))


def main(args=None):
    """The main function called by the `fitsinfo` script."""
    parser = argparse.ArgumentParser(
        description=DESCRIPTION, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    # TODO: pass suggest_on_error as kwarg when PYTHON_LT_14 is dropped
    parser.suggest_on_error = True

    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "filename",
        nargs="+",
        help="Path to one or more FITS files. Wildcards are supported.",
    )
    args = parser.parse_args(args)

    for idx, filename in enumerate(args.filename):
        if idx > 0:
            print()
        fitsinfo(filename)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

from astropy.cosmology._src.io.builtin.latex import _FORMAT_TABLE, write_latex
from astropy.io.registry.base import IORegistryError
from astropy.table import QTable, Table

from .base import ReadWriteDirectTestBase, ReadWriteTestMixinBase


class WriteLATEXTestMixin(ReadWriteTestMixinBase):
    """
    Tests for a Cosmology[Write] with ``format="latex"``.
    This class will not be directly called by :mod:`pytest` since its name does
    not begin with ``Test``. To activate the contained tests this class must
    be inherited in a subclass. Subclasses must define a :func:`pytest.fixture`
    ``cosmo`` that returns/yields an instance of a |Cosmology|.
    See ``TestCosmology`` for an example.
    """

    def test_to_latex_failed_cls(self, write, tmp_path):
        """Test failed table type."""
        fp = tmp_path / "test_to_latex_failed_cls.tex"

        with pytest.raises(TypeError, match="'cls' must be"):
            write(fp, cls=list)

    @pytest.mark.parametrize("tbl_cls", [QTable, Table])
    def test_to_latex_cls(self, write, tbl_cls, tmp_path):
        fp = tmp_path / "test_to_latex_cls.tex"
        write(fp, cls=tbl_cls)

    def test_latex_columns(self, write, tmp_path):
        fp = tmp_path / "test_rename_latex_columns.tex"
        write(fp, latex_names=True)
        tbl = QTable.read(fp)
        # asserts each column name has not been reverted yet
        # For now, Cosmology class and name are stored in first 2 slots
        for column_name in tbl.colnames[2:]:
            assert column_name in _FORMAT_TABLE.values()

    def test_write_latex_invalid_path(self, write):
        """Test passing an invalid path"""
        invalid_fp = ""
        with pytest.raises(FileNotFoundError, match="No such file or directory"):
            write(invalid_fp, format="ascii.latex")

    def test_write_latex_false_overwrite(self, write, tmp_path):
        """Test to write a LaTeX file without overwriting an existing file"""
        # Test that passing an invalid path to write_latex() raises a IOError
        fp = tmp_path / "test_write_latex_false_overwrite.tex"
        write(fp)
        with pytest.raises(OSError, match="overwrite=True"):
            write(fp, overwrite=False)

    def test_write_latex_unsupported_format(self, write, tmp_path):
        """Test for unsupported format"""
        fp = tmp_path / "test_write_latex_unsupported_format.tex"
        invalid_format = "unsupported"
        with pytest.raises((ValueError, IORegistryError)) as exc_info:
            pytest.raises(ValueError, match="format must be 'ascii.latex'")
            pytest.raises(IORegistryError, match="No writer defined for format")
            write(fp, format=invalid_format)


class TestReadWriteLaTex(ReadWriteDirectTestBase, WriteLATEXTestMixin):
    """
    Directly test ``write_latex``.
    These are not public API and are discouraged from use, in favor of
    ``Cosmology.write(..., format="latex")``, but should be
    tested regardless b/c they are used internally.
    """

    def setup_class(self):
        self.functions = {"write": write_latex}

    def test_rename_direct_latex_columns(self, write, tmp_path):
        """Tests renaming columns"""
        fp = tmp_path / "test_rename_latex_columns.tex"
        write(fp, latex_names=True)
        tbl = QTable.read(fp)
        # asserts each column name has not been reverted yet
        for column_name in tbl.colnames[2:]:
            # for now, Cosmology as metadata and name is stored in first 2 slots
            assert column_name in _FORMAT_TABLE.values()



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Testing :mod:`astropy.cosmology.flrw.base`.

This module sets up the tests for subclasses of :class:`astropy.cosmology.FLRW`. The
tests for the specific abstract class :class:`astropy.cosmology.FLRW` are in
``test_flrw``.

"""

import abc
from functools import cached_property

import numpy as np
import pytest

import astropy.constants as const
import astropy.units as u
from astropy.cosmology import FLRW, FlatLambdaCDM, LambdaCDM, Planck18
from astropy.cosmology._src.core import _COSMOLOGY_CLASSES, dataclass_decorator
from astropy.cosmology._src.flrw.base import a_B_c2
from astropy.cosmology._src.tests.helper import get_redshift_methods
from astropy.cosmology._src.tests.test_core import (
    CosmologyTest,
    FlatCosmologyMixinTest,
    invalid_zs,
    valid_zs,
)
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.compat.optional_deps import HAS_PANDAS, HAS_SCIPY

from .conftest import filter_keys_from_items
from .test_parameters import (
    ParameterFlatOde0TestMixin,
    ParameterH0TestMixin,
    Parameterm_nuTestMixin,
    ParameterNeffTestMixin,
    ParameterOb0TestMixin,
    ParameterOde0TestMixin,
    ParameterOm0TestMixin,
    ParameterTcmb0TestMixin,
)

##############################################################################
# TESTS
##############################################################################


class FLRWTest(
    CosmologyTest,
    ParameterH0TestMixin,
    ParameterOm0TestMixin,
    ParameterOde0TestMixin,
    ParameterTcmb0TestMixin,
    ParameterNeffTestMixin,
    Parameterm_nuTestMixin,
    ParameterOb0TestMixin,
):
    abstract_w = False

    @abc.abstractmethod
    def setup_class(self):
        """Setup for testing."""
        super().setup_class(self)

        # Default cosmology args and kwargs
        self._cls_args = dict(
            H0=70 * u.km / u.s / u.Mpc, Om0=0.27 * u.one, Ode0=0.73 * u.one
        )
        self.cls_kwargs = dict(
            Tcmb0=3.0 * u.K,
            Ob0=0.03 * u.one,
            name=self.__class__.__name__,
            meta={"a": "b"},
        )

    @pytest.fixture(scope="class")
    def nonflatcosmo(self):
        """A non-flat cosmology used in equivalence tests."""
        return LambdaCDM(70, 0.4, 0.8)

    # ===============================================================
    # Method & Attribute Tests

    def test_init(self, cosmo_cls):
        """Test initialization."""
        super().test_init(cosmo_cls)

        # TODO! tests for initializing calculated values, e.g. `h`
        # TODO! transfer tests for initializing neutrinos

    def test_init_Tcmb0_zeroing(self, cosmo_cls, ba):
        """Test if setting Tcmb0 parameter to 0 influences other parameters.

        TODO: consider moving this test to ``FLRWTest``
        """
        ba.arguments["Tcmb0"] = 0.0
        cosmo = cosmo_cls(*ba.args, **ba.kwargs)

        assert cosmo.Ogamma0 == 0.0
        assert cosmo.Onu0 == 0.0

        if not self.abstract_w:
            assert u.allclose(cosmo.Ogamma(1.5), [0, 0, 0, 0])
            assert u.allclose(cosmo.Ogamma([0, 1, 2, 3]), [0, 0, 0, 0])
            assert u.allclose(cosmo.Onu(1.5), [0, 0, 0, 0])
            assert u.allclose(cosmo.Onu([0, 1, 2, 3]), [0, 0, 0, 0])

    # ---------------------------------------------------------------
    # Properties

    def test_Odm0(self, cosmo_cls, cosmo):
        """Test ``cached_property`` ``Odm0``."""
        # on the class
        assert isinstance(cosmo_cls.Odm0, cached_property)

        # on the instance
        assert np.allclose(cosmo.Odm0, cosmo.Om0 - cosmo.Ob0)

    def test_Ok0(self, cosmo_cls, cosmo):
        """Test ``cached_property`` ``Ok0``."""
        # on the class
        assert isinstance(cosmo_cls.Ok0, cached_property)

        # on the instance
        assert np.allclose(
            cosmo.Ok0, 1.0 - (cosmo.Om0 + cosmo.Ode0 + cosmo.Ogamma0 + cosmo.Onu0)
        )

    def test_is_flat(self, cosmo_cls, cosmo):
        """Test property ``is_flat``."""
        # on the class
        assert isinstance(cosmo_cls.is_flat, property)
        assert cosmo_cls.is_flat.fset is None  # immutable

        # on the instance
        assert isinstance(cosmo.is_flat, bool)
        assert cosmo.is_flat is bool((cosmo.Ok0 == 0.0) and (cosmo.Otot0 == 1.0))

    def test_Tnu0(self, cosmo_cls, cosmo):
        """Test ``cached_property`` ``Tnu0``."""
        # on the class
        assert isinstance(cosmo_cls.Tnu0, cached_property)

        # on the instance
        assert cosmo.Tnu0.unit == u.K
        assert u.allclose(cosmo.Tnu0, 0.7137658555036082 * cosmo.Tcmb0, rtol=1e-5)

    def test_has_massive_nu(self, cosmo_cls, cosmo):
        """Test property ``has_massive_nu``."""
        # on the class
        assert isinstance(cosmo_cls.has_massive_nu, property)
        assert cosmo_cls.has_massive_nu.fset is None  # immutable

        # on the instance
        if cosmo.Tnu0 == 0:
            assert cosmo.has_massive_nu is False
        else:
            assert cosmo.has_massive_nu is cosmo._nu_info.has_massive_nu

    def test_h(self, cosmo_cls, cosmo):
        """Test ``cached_property`` ``h``."""
        # on the class
        assert isinstance(cosmo_cls.h, cached_property)

        # on the instance
        assert np.allclose(cosmo.h, cosmo.H0.value / 100.0)

    def test_hubble_time(self, cosmo_cls, cosmo):
        """Test ``cached_property`` ``hubble_time``."""
        # on the class
        assert isinstance(cosmo_cls.hubble_time, cached_property)

        # on the instance
        assert u.allclose(cosmo.hubble_time, (1 / cosmo.H0) << u.Gyr)

    def test_hubble_distance(self, cosmo_cls, cosmo):
        """Test ``cached_property`` ``hubble_distance``."""
        # on the class
        assert isinstance(cosmo_cls.hubble_distance, cached_property)

        # on the instance
        assert cosmo.hubble_distance == (const.c / cosmo.H0).to(u.Mpc)

    def test_critical_density0(self, cosmo_cls, cosmo):
        """Test ``cached_property`` ``critical_density0``."""
        # on the class
        assert isinstance(cosmo_cls.critical_density0, cached_property)

        # on the instance
        assert cosmo.critical_density0.unit == u.g / u.cm**3
        assert u.allclose(  # sanity check
            cosmo.critical_density0, 3 * cosmo.H0**2 / (8 * np.pi * const.G)
        )

    def test_Ogamma0(self, cosmo_cls, cosmo):
        """Test ``cached_property`` ``Ogamma0``."""
        # on the class
        assert isinstance(cosmo_cls.Ogamma0, cached_property)

        # on the instance
        # Ogamma cor \propto T^4/rhocrit
        expect = a_B_c2 * cosmo.Tcmb0.value**4 / cosmo.critical_density0.value
        assert np.allclose(cosmo.Ogamma0, expect)
        # check absolute equality to 0 if Tcmb0 is 0
        if cosmo.Tcmb0 == 0:
            assert cosmo.Ogamma0 == 0

    def test_Onu0(self, cosmo_cls, cosmo):
        """Test ``cached_property`` ``Onu0``."""
        # on the class
        assert isinstance(cosmo_cls.Onu0, cached_property)

        # on the instance
        # neutrino temperature <= photon temperature since the neutrinos
        # decouple first.
        if cosmo.has_massive_nu:  # Tcmb0 > 0 & has massive
            # check the expected formula
            assert cosmo.Onu0 == cosmo.Ogamma0 * cosmo.nu_relative_density(0)
            # a sanity check on on the ratio of neutrinos to photons
            # technically it could be 1, but not for any of the tested cases.
            assert cosmo.nu_relative_density(0) <= 1
        elif cosmo.Tcmb0 == 0:
            assert cosmo.Onu0 == 0
        else:
            # check the expected formula
            assert cosmo.Onu0 == 0.22710731766 * cosmo.__dict__["Neff"] * cosmo.Ogamma0
            # and check compatibility with nu_relative_density
            assert np.allclose(
                cosmo.nu_relative_density(0), 0.22710731766 * cosmo.__dict__["Neff"]
            )

    def test_Otot0(self, cosmo):
        """Test :attr:`astropy.cosmology.FLRW.Otot0`."""
        assert (
            cosmo.Otot0
            == cosmo.Om0 + cosmo.Ogamma0 + cosmo.Onu0 + cosmo.Ode0 + cosmo.Ok0
        )

    # ---------------------------------------------------------------
    # Methods

    _FLRW_redshift_methods = get_redshift_methods(
        FLRW, include_private=True, include_z2=False
    )

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy is not installed")
    @pytest.mark.parametrize("z, exc", invalid_zs)
    @pytest.mark.parametrize("method", sorted(_FLRW_redshift_methods))
    def test_redshift_method_bad_input(self, cosmo, method, z, exc):
        """Test all the redshift methods for bad input."""
        with pytest.raises(exc):
            getattr(cosmo, method)(z)

    @pytest.mark.parametrize("z", valid_zs)
    @abc.abstractmethod
    def test_w(self, cosmo, z):
        """Test :meth:`astropy.cosmology.FLRW.w`.

        Since ``w`` is abstract, each test class needs to define further tests.
        """
        # super().test_w(cosmo, z)  # NOT b/c abstract `w(z)`
        w = cosmo.w(z)
        assert np.shape(w) == np.shape(z)  # test same shape
        assert u.Quantity(w).unit == u.one  # test no units or dimensionless

    # -------------------------------------------

    @pytest.mark.parametrize("z", valid_zs)
    def test_Otot(self, cosmo, z):
        """Test :meth:`astropy.cosmology.FLRW.Otot`."""
        # super().test_Otot(cosmo)  # NOT b/c abstract `w(z)`
        assert np.allclose(
            cosmo.Otot(z),
            cosmo.Om(z) + cosmo.Ogamma(z) + cosmo.Onu(z) + cosmo.Ode(z) + cosmo.Ok(z),
        )

    def test_scale_factor0(self, cosmo):
        """Test :meth:`astropy.cosmology.FLRW.scale_factor`."""
        assert isinstance(cosmo.scale_factor0, u.Quantity)
        assert cosmo.scale_factor0.unit == u.one
        assert cosmo.scale_factor0 == 1
        assert np.allclose(cosmo.scale_factor0, cosmo.scale_factor(0))

    @pytest.mark.parametrize("z", valid_zs)
    def test_scale_factor(self, cosmo, z):
        """Test :meth:`astropy.cosmology.FLRW.scale_factor`."""
        assert np.allclose(cosmo.scale_factor(z), 1 / (1 + np.array(z)))

    # -------------------------------------------

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy required for this test.")
    def test_comoving_distance_1arg_equal_to_2arg(self, cosmo):
        """Test :meth:`astropy.cosmology.FLRW.comoving_distance`."""
        # Special case of z1 = 0
        z = np.linspace(0, 1, 10)
        assert u.allclose(cosmo.comoving_distance(z), cosmo.comoving_distance(0, z))

        # General case of z1, z2
        z1 = z
        z2 = z + 1
        assert u.allclose(
            cosmo.comoving_distance(z2) - cosmo.comoving_distance(z1),
            cosmo.comoving_distance(z1, z2),
        )

    @pytest.mark.skipif(
        not (HAS_PANDAS and HAS_SCIPY), reason="requires pandas and scipy"
    )
    def test_luminosity_distance_pandas(self, cosmo):
        """Test :meth:`astropy.cosmology.FLRW.luminosity_distance`.

        Regression test for https://github.com/astropy/astropy/issues/15576.
        """
        import pandas as pd

        z = pd.Series([0.1, 0.2, 0.3])
        d = cosmo.luminosity_distance(z)

        assert isinstance(d, u.Quantity)
        assert d.unit == u.Mpc
        np.testing.assert_array_equal(d, cosmo.luminosity_distance(np.array(z)))

    # ---------------------------------------------------------------

    def test_efunc_vs_invefunc(self, cosmo):
        """Test that ``efunc`` and ``inv_efunc`` give inverse values.

        Note that the test doesn't need scipy because it doesn't need to call
        ``de_density_scale``.
        """
        # super().test_efunc_vs_invefunc(cosmo)  # NOT b/c abstract `w(z)`
        z0 = 0.5
        z = np.array([0.5, 1.0, 2.0, 5.0])

        assert np.allclose(cosmo.efunc(z0), 1.0 / cosmo.inv_efunc(z0))
        assert np.allclose(cosmo.efunc(z), 1.0 / cosmo.inv_efunc(z))

    # ---------------------------------------------------------------
    # from Cosmology

    def test_clone_change_param(self, cosmo):
        """Test method ``.clone()`` changing a(many) Parameter(s)."""
        super().test_clone_change_param(cosmo)

        # don't change any values
        kwargs = dict(cosmo.parameters)
        c = cosmo.clone(**kwargs)
        assert c.__class__ == cosmo.__class__
        assert c == cosmo

        # change ``H0``
        # Note that H0 affects Ode0 because it changes Ogamma0
        c = cosmo.clone(H0=100)
        assert c.__class__ == cosmo.__class__
        assert c.name == cosmo.name + " (modified)"
        assert c.H0.value == 100
        for n, v in filter_keys_from_items(c.parameters, ("H0",)):
            v_expect = getattr(cosmo, n)
            assert_quantity_allclose(v, v_expect, atol=1e-4 * getattr(v, "unit", 1))
        assert not u.allclose(c.Ogamma0, cosmo.Ogamma0)
        assert not u.allclose(c.Onu0, cosmo.Onu0)

        # change multiple things
        c = cosmo.clone(name="new name", H0=100, Tcmb0=2.8, meta=dict(zz="tops"))
        assert c.__class__ == cosmo.__class__
        assert c.name == "new name"
        assert c.H0.value == 100
        assert c.Tcmb0.value == 2.8
        assert c.meta == {**cosmo.meta, **dict(zz="tops")}
        for n, v in filter_keys_from_items(c.parameters, ("H0", "Tcmb0")):
            v_expect = getattr(cosmo, n)
            assert_quantity_allclose(v, v_expect, atol=1e-4 * getattr(v, "unit", 1))
        assert not u.allclose(c.Ogamma0, cosmo.Ogamma0)
        assert not u.allclose(c.Onu0, cosmo.Onu0)
        assert not u.allclose(c.Tcmb0.value, cosmo.Tcmb0.value)

    def test_is_equivalent(self, cosmo):
        """Test :meth:`astropy.cosmology.FLRW.is_equivalent`."""
        super().test_is_equivalent(cosmo)  # pass to CosmologyTest

        # test against a FlatFLRWMixin
        # case (3) in FLRW.is_equivalent
        if isinstance(cosmo, FlatLambdaCDM):
            assert cosmo.is_equivalent(Planck18)
            assert Planck18.is_equivalent(cosmo)
        else:
            assert not cosmo.is_equivalent(Planck18)
            assert not Planck18.is_equivalent(cosmo)

    # ===============================================================
    # Usage Tests

    # TODO: this test should be subsumed by other tests
    @pytest.mark.parametrize("method", ("Om", "Ode", "w", "de_density_scale"))
    def test_distance_broadcast(self, cosmo, method):
        """Test distance methods broadcast z correctly."""
        g = getattr(cosmo, method)
        z = np.linspace(0.1, 1, 6)
        z2d = z.reshape(2, 3)
        z3d = z.reshape(3, 2, 1)

        value_flat = g(z)
        assert value_flat.shape == z.shape

        value_2d = g(z2d)
        assert value_2d.shape == z2d.shape

        value_3d = g(z3d)
        assert value_3d.shape == z3d.shape
        assert u.allclose(value_flat, value_2d.flatten())
        assert u.allclose(value_flat, value_3d.flatten())

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy required for this test.")
    def test_comoving_distance_example(self, cosmo_cls, args, kwargs, expected):
        """Test :meth:`astropy.cosmology.LambdaCDM.comoving_distance`.

        These do not come from external codes -- they are just internal checks to make
        sure nothing changes if we muck with the distance calculators.
        """
        z = np.array([1.0, 2.0, 3.0, 4.0])

        cosmo = cosmo_cls(*args, **kwargs)
        assert u.allclose(cosmo.comoving_distance(z), expected, rtol=1e-4)


# ==============================================================================


class FlatFLRWMixinTest(FlatCosmologyMixinTest, ParameterFlatOde0TestMixin):
    """Tests for :class:`astropy.cosmology.FlatFLRWMixin` subclasses.

    E.g to use this class::

        class TestFlatSomeFLRW(FlatFLRWMixinTest, TestSomeFLRW):
            ...
    """

    def setup_class(self):
        """Setup for testing.

        Set up as for regular FLRW test class, but remove dark energy component
        since flat cosmologies are forbidden Ode0 as an argument,
        see ``test_init_subclass``.
        """
        super().setup_class(self)
        self._cls_args.pop("Ode0")

    # ===============================================================
    # Method & Attribute Tests

    # ---------------------------------------------------------------
    # class-level

    def test_init_subclass(self, cosmo_cls):
        """Test initializing subclass, mostly that can't have Ode0 in init."""
        super().test_init_subclass(cosmo_cls)

        with pytest.raises(TypeError, match="subclasses of"):

            @dataclass_decorator
            class HASOde0SubClass(cosmo_cls):
                def __init__(self, Ode0):
                    pass

            _COSMOLOGY_CLASSES.pop(HASOde0SubClass.__qualname__, None)

    # ---------------------------------------------------------------
    # instance-level

    def test_init(self, cosmo_cls):
        super().test_init(cosmo_cls)

        cosmo = cosmo_cls(*self.cls_args, **self.cls_kwargs)
        assert cosmo.Ok0 == 0.0
        assert cosmo.Ode0 == 1.0 - (cosmo.Om0 + cosmo.Ogamma0 + cosmo.Onu0 + cosmo.Ok0)

    def test_Ok0(self, cosmo_cls, cosmo):
        """Test property ``Ok0``."""
        super().test_Ok0(cosmo_cls, cosmo)

        # for flat cosmologies, Ok0 is not *close* to 0, it *is* 0
        assert cosmo.Ok0 == 0.0

    def test_Otot0(self, cosmo):
        """Test :attr:`astropy.cosmology.FLRW.Otot0`. Should always be 1."""
        super().test_Otot0(cosmo)

        # for flat cosmologies, Otot0 is not *close* to 1, it *is* 1
        assert cosmo.Otot0 == 1.0

    @pytest.mark.parametrize("z", valid_zs)
    def test_Otot(self, cosmo, z):
        """Test :meth:`astropy.cosmology.FLRW.Otot`. Should always be 1."""
        super().test_Otot(cosmo, z)

        # for flat cosmologies, Otot is 1, within precision.
        assert u.allclose(cosmo.Otot(z), 1.0)

    @pytest.mark.skipif(not HAS_SCIPY, reason="scipy is not installed")
    @pytest.mark.parametrize("z, exc", invalid_zs)
    @pytest.mark.parametrize(
        "method", sorted(FLRWTest._FLRW_redshift_methods - {"Otot"})
    )
    def test_redshift_method_bad_input(self, cosmo, method, z, exc):
        """Test all the redshift methods for bad input."""
        super().test_redshift_method_bad_input(cosmo, method, z, exc)

    # ---------------------------------------------------------------

    def test_clone_to_nonflat_change_param(self, cosmo):
        """Test method ``.clone()`` changing a(many) Parameter(s)."""
        super().test_clone_to_nonflat_change_param(cosmo)

        # change Ode0, without non-flat
        msg = "Cannot set 'Ode0' in clone unless 'to_nonflat=True'. "
        with pytest.raises(ValueError, match=msg):
            cosmo.clone(Ode0=1)

        # change to non-flat
        nc = cosmo.clone(to_nonflat=True, Ode0=cosmo.Ode0)
        assert isinstance(nc, cosmo.__nonflatclass__)
        assert nc == cosmo.nonflat

        nc = cosmo.clone(to_nonflat=True, Ode0=1)
        assert nc.Ode0 == 1.0
        assert nc.name == cosmo.name + " (modified)"

    # ---------------------------------------------------------------

    def test_is_equivalent(self, cosmo, nonflatcosmo):
        """Test :meth:`astropy.cosmology.FLRW.is_equivalent`."""
        super().test_is_equivalent(cosmo)  # pass to TestFLRW

        # against non-flat Cosmology
        assert not cosmo.is_equivalent(nonflatcosmo)
        assert not nonflatcosmo.is_equivalent(cosmo)

        # non-flat version of class
        nonflat_cosmo_cls = cosmo.__nonflatclass__
        # keys check in `test_is_equivalent_nonflat_class_different_params`

        # non-flat
        nonflat = nonflat_cosmo_cls(*self.cls_args, Ode0=0.9, **self.cls_kwargs)
        assert not nonflat.is_equivalent(cosmo)
        assert not cosmo.is_equivalent(nonflat)

        # Flat, but not FlatFLRWMixin
        # This will require forcing flatness by overriding attribute values.
        # Since Cosmology is frozen, the easiest way is via __dict__.
        flat = nonflat_cosmo_cls(
            *self.cls_args,
            Ode0=1.0 - cosmo.Om0 - cosmo.Ogamma0 - cosmo.Onu0,
            **self.cls_kwargs,
        )
        flat.__dict__["Ok0"] = 0.0  # manually forcing flatness by setting `Ok0`.
        assert flat.is_equivalent(cosmo)
        assert cosmo.is_equivalent(flat)

    def test_repr(self, cosmo_cls, cosmo):
        """
        Test method ``.__repr__()``. Skip non-flat superclass test.
        e.g. `TestFlatLambdaCDDM` -> `FlatFLRWMixinTest`
        vs   `TestFlatLambdaCDDM` -> `TestLambdaCDDM` -> `FlatFLRWMixinTest`
        """
        # test eliminated Ode0 from parameters
        assert "Ode0" not in repr(cosmo)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Test importability of common Masked classes."""

import pytest

from astropy.coordinates.angles.core import Angle, Latitude, Longitude
from astropy.units.quantity import Quantity
from astropy.utils.masked import core


@pytest.mark.parametrize("base_class", [Quantity, Angle, Latitude, Longitude])
def test_importable(base_class):
    assert getattr(core, f"Masked{base_class.__name__}") is core.Masked(base_class)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module provides utility functions for the models package.
"""

import warnings

# pylint: disable=invalid-name
from collections import UserDict
from collections.abc import Sequence
from inspect import signature
from typing import TypeVar, overload

import numpy as np
from numpy.typing import NDArray

from astropy import units as u

__all__ = ["ellipse_extent", "poly_map_domain"]


def make_binary_operator_eval(oper, f, g):
    """
    Given a binary operator (as a callable of two arguments) ``oper`` and
    two callables ``f`` and ``g`` which accept the same arguments,
    returns a *new* function that takes the same arguments as ``f`` and ``g``,
    but passes the outputs of ``f`` and ``g`` in the given ``oper``.

    ``f`` and ``g`` are assumed to return tuples (which may be 1-tuples).  The
    given operator is applied element-wise to tuple outputs).

    Example
    -------
    >>> from operator import add
    >>> def prod(x, y):
    ...     return (x * y,)
    ...
    >>> sum_of_prod = make_binary_operator_eval(add, prod, prod)
    >>> sum_of_prod(3, 5)
    (30,)
    """
    return lambda inputs, params: tuple(
        oper(x, y) for x, y in zip(f(inputs, params), g(inputs, params))
    )


def poly_map_domain(oldx, domain, window):
    """
    Map domain into window by shifting and scaling.

    Parameters
    ----------
    oldx : array
          original coordinates
    domain : list or tuple of length 2
          function domain
    window : list or tuple of length 2
          range into which to map the domain
    """
    domain = np.array(domain, dtype=np.float64)
    window = np.array(window, dtype=np.float64)
    if domain.shape != (2,) or window.shape != (2,):
        raise ValueError('Expected "domain" and "window" to be a tuple of size 2.')
    scl = (window[1] - window[0]) / (domain[1] - domain[0])
    off = (window[0] * domain[1] - window[1] * domain[0]) / (domain[1] - domain[0])
    return off + scl * oldx


def _validate_domain_window(value):
    if value is not None:
        if np.asanyarray(value).shape != (2,):
            raise ValueError("domain and window should be tuples of size 2.")
        return tuple(value)
    return value


def array_repr_oneline(array):
    """
    Represents a multi-dimensional Numpy array flattened onto a single line.
    """
    r = np.array2string(array, separator=", ", suppress_small=True)
    return " ".join(line.strip() for line in r.splitlines())


def combine_labels(left, right):
    """
    For use with the join operator &: Combine left input/output labels with
    right input/output labels.

    If none of the labels conflict then this just returns a sum of tuples.
    However if *any* of the labels conflict, this appends '0' to the left-hand
    labels and '1' to the right-hand labels so there is no ambiguity).
    """
    if set(left).intersection(right):
        left = tuple(label + "0" for label in left)
        right = tuple(label + "1" for label in right)

    return left + right


def ellipse_extent(a, b, theta):
    """
    Calculates the half size of a box encapsulating a rotated 2D
    ellipse.

    Parameters
    ----------
    a : float or `~astropy.units.Quantity`
        The ellipse semimajor axis.
    b : float or `~astropy.units.Quantity`
        The ellipse semiminor axis.
    theta : float or `~astropy.units.Quantity` ['angle']
        The rotation angle as an angular quantity
        (`~astropy.units.Quantity` or `~astropy.coordinates.Angle`) or
        a value in radians (as a float). The rotation angle increases
        counterclockwise.

    Returns
    -------
    offsets : tuple
        The absolute value of the offset distances from the ellipse center that
        define its bounding box region, ``(dx, dy)``.

    Examples
    --------
    .. plot::
        :include-source:

        import numpy as np
        import matplotlib.pyplot as plt
        from astropy.modeling.models import Ellipse2D
        from astropy.modeling.utils import ellipse_extent, render_model

        amplitude = 1
        x0 = 50
        y0 = 50
        a = 30
        b = 10
        theta = np.pi / 4

        model = Ellipse2D(amplitude, x0, y0, a, b, theta)
        dx, dy = ellipse_extent(a, b, theta)
        limits = [x0 - dx, x0 + dx, y0 - dy, y0 + dy]
        model.bounding_box = limits

        image = render_model(model)

        plt.imshow(image, cmap='binary', interpolation='nearest', alpha=.5,
                  extent = limits)
        plt.show()
    """
    from .parameters import Parameter  # prevent circular import

    if isinstance(theta, Parameter):
        if theta.quantity is None:
            theta = theta.value
        else:
            theta = theta.quantity

    t = np.arctan2(-b * np.tan(theta), a)
    dx = a * np.cos(t) * np.cos(theta) - b * np.sin(t) * np.sin(theta)

    t = np.arctan2(b, a * np.tan(theta))
    dy = b * np.sin(t) * np.cos(theta) + a * np.cos(t) * np.sin(theta)

    if isinstance(dx, u.Quantity) or isinstance(dy, u.Quantity):
        return np.abs(u.Quantity([dx, dy], subok=True))
    return np.abs([dx, dy])


def get_inputs_and_params(func):
    """
    Given a callable, determine the input variables and the
    parameters.

    Parameters
    ----------
    func : callable

    Returns
    -------
    inputs, params : tuple
        Each entry is a list of inspect.Parameter objects
    """
    sig = signature(func)

    inputs = []
    params = []
    for param in sig.parameters.values():
        if param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
            raise ValueError("Signature must not have *args or **kwargs")
        if param.default == param.empty:
            inputs.append(param)
        else:
            params.append(param)

    return inputs, params


def _combine_equivalency_dict(keys, eq1=None, eq2=None):
    # Given two dictionaries that give equivalencies for a set of keys, for
    # example input value names, return a dictionary that includes all the
    # equivalencies
    eq = {}
    for key in keys:
        eq[key] = []
        if eq1 is not None and key in eq1:
            eq[key].extend(eq1[key])
        if eq2 is not None and key in eq2:
            eq[key].extend(eq2[key])
    return eq


def _to_radian(value):
    """Convert ``value`` to radian."""
    if isinstance(value, u.Quantity):
        return value.to(u.rad)
    return np.deg2rad(value)


def _to_orig_unit(value, raw_unit=None, orig_unit=None):
    """Convert value with ``raw_unit`` to ``orig_unit``."""
    if raw_unit is not None:
        return (value * raw_unit).to(orig_unit)
    return np.rad2deg(value)


class _ConstraintsDict(UserDict):
    """
    Wrapper around UserDict to allow updating the constraints
    on a Parameter when the dictionary is updated.
    """

    def __init__(self, model, constraint_type):
        self._model = model
        self.constraint_type = constraint_type
        c = {}
        for name in model.param_names:
            param = getattr(model, name)
            c[name] = getattr(param, constraint_type)
        super().__init__(c)

    def __setitem__(self, key, val):
        super().__setitem__(key, val)
        param = getattr(self._model, key)
        setattr(param, self.constraint_type, val)


class _SpecialOperatorsDict(UserDict):
    """
    Wrapper around UserDict to allow for better tracking of the Special
    Operators for CompoundModels. This dictionary is structured so that
    one cannot inadvertently overwrite an existing special operator.

    Parameters
    ----------
    unique_id: int
        the last used unique_id for a SPECIAL OPERATOR
    special_operators: dict
        a dictionary containing the special_operators

    Notes
    -----
    Direct setting of operators (`dict[key] = value`) into the
    dictionary has been deprecated in favor of the `.add(name, value)`
    method, so that unique dictionary keys can be generated and tracked
    consistently.
    """

    def __init__(self, unique_id=0, special_operators={}):
        super().__init__(special_operators)
        self._unique_id = unique_id

    def _set_value(self, key, val):
        if key in self:
            raise ValueError(f'Special operator "{key}" already exists')
        else:
            super().__setitem__(key, val)

    def __setitem__(self, key, val):
        self._set_value(key, val)
        warnings.warn(
            DeprecationWarning(
                """
            Special operator dictionary assignment has been deprecated.
            Please use `.add` instead, so that you can capture a unique
            key for your operator.
            """
            )
        )

    def _get_unique_id(self):
        self._unique_id += 1

        return self._unique_id

    def add(self, operator_name, operator):
        """
        Adds a special operator to the dictionary, and then returns the
        unique key that the operator is stored under for later reference.

        Parameters
        ----------
        operator_name: str
            the name for the operator
        operator: function
            the actual operator function which will be used

        Returns
        -------
        the unique operator key for the dictionary
            `(operator_name, unique_id)`
        """
        key = (operator_name, self._get_unique_id())

        self._set_value(key, operator)

        return key


DType = TypeVar("DType", bound=np.generic)


@overload
def quantity_asanyarray(a: Sequence[int]) -> NDArray[np.integer]: ...
@overload
def quantity_asanyarray(a: Sequence[int], dtype: DType) -> NDArray[DType]: ...
@overload
def quantity_asanyarray(a: Sequence[u.Quantity]) -> u.Quantity: ...
def quantity_asanyarray(
    a: Sequence[int] | Sequence[u.Quantity], dtype: DType | None = None
) -> NDArray[np.integer] | NDArray[DType] | u.Quantity:
    if (
        not isinstance(a, np.ndarray)
        and not np.isscalar(a)
        and any(isinstance(x, u.Quantity) for x in a)
    ):
        return u.Quantity(a, dtype=dtype)
    else:
        # skip over some dtype deprecation.
        dtype = np.float64 if dtype is np.inexact else dtype
        return np.asanyarray(a, dtype=dtype)



# Licensed under a 3-clause BSD style license - see LICNSE.rst

# This module includes files automatically generated from ply (these end in
# _lextab.py and _parsetab.py). To generate these files, remove them from this
# folder, then build astropy and run the tests in-place:
#
#   python setup.py build_ext --inplace
#   pytest astropy/units
#
# You can then commit the changes to the re-generated _lextab.py and
# _parsetab.py files.

"""Handles the CDS string format for units."""

import re
from typing import ClassVar, Literal

from astropy.extern.ply.lex import Lexer
from astropy.units.core import CompositeUnit, Unit, UnitBase
from astropy.units.enums import DeprecatedUnitAction
from astropy.units.utils import is_effectively_unity
from astropy.utils import classproperty, parsing
from astropy.utils.parsing import ThreadSafeParser

from .base import Base, _ParsingFormatMixin


class CDS(Base, _ParsingFormatMixin):
    """
    Support the `Centre de Données astronomiques de Strasbourg
    <https://cds.unistra.fr/>`_ `Standards for Astronomical
    Catalogues 2.0 <https://vizier.unistra.fr/vizier/doc/catstd-3.2.htx>`_
    format, and the `complete set of supported units
    <https://vizier.unistra.fr/viz-bin/Unit>`_.  This format is used
    by VOTable up to version 1.2.
    """

    _space: ClassVar[str] = "."
    _times: ClassVar[str] = "x"
    _scale_unit_separator: ClassVar[str] = ""

    _tokens: ClassVar[tuple[str, ...]] = (
        "PRODUCT",
        "DIVISION",
        "OPEN_PAREN",
        "CLOSE_PAREN",
        "OPEN_BRACKET",
        "CLOSE_BRACKET",
        "X",
        "SIGN",
        "UINT",
        "UFLOAT",
        "UNIT",
        "DIMENSIONLESS",
    )

    @classproperty(lazy=True)
    def _units(cls) -> dict[str, UnitBase]:
        from astropy import units as u
        from astropy.units import cds

        return {k: v for k, v in cds.__dict__.items() if isinstance(v, u.UnitBase)}

    @classproperty(lazy=True)
    def _lexer(cls) -> Lexer:
        tokens = cls._tokens

        t_PRODUCT = r"\."
        t_DIVISION = r"/"
        t_OPEN_PAREN = r"\("
        t_CLOSE_PAREN = r"\)"
        t_OPEN_BRACKET = r"\["
        t_CLOSE_BRACKET = r"\]"

        # NOTE THE ORDERING OF THESE RULES IS IMPORTANT!!
        # Regular expression rules for simple tokens

        def t_UFLOAT(t):
            r"((\d+\.?\d+)|(\.\d+))([eE][+-]?\d+)?"
            if not re.search(r"[eE\.]", t.value):
                t.type = "UINT"
                t.value = int(t.value)
            else:
                t.value = float(t.value)
            return t

        def t_UINT(t):
            r"\d+"
            t.value = int(t.value)
            return t

        def t_SIGN(t):
            r"[+-](?=\d)"
            t.value = float(t.value + "1")
            return t

        def t_X(t):  # multiplication for factor in front of unit
            r"[x×]"
            return t

        # Most units are just combinations of letters with no numbers, but there
        # are a few special ones (\h is Planch constant) and three that end in 0.
        def t_UNIT(t):
            r"%|°|\\h|(a|eps|mu)0|((?!\d)\w)+"
            t.value = cls._get_unit(t)
            return t

        def t_DIMENSIONLESS(t):
            r"---|-"
            # These are separate from t_UNIT since they cannot have a prefactor.
            t.value = cls._get_unit(t)
            return t

        t_ignore = ""

        # Error handling rule
        def t_error(t):
            raise ValueError(f"Invalid character at col {t.lexpos}")

        return parsing.lex(
            lextab="cds_lextab", package="astropy/units", reflags=int(re.UNICODE)
        )

    @classproperty(lazy=True)
    def _parser(cls) -> ThreadSafeParser:
        """
        The grammar here is based on the description in the `Standards
        for Astronomical Catalogues 2.0
        <https://vizier.unistra.fr/vizier/doc/catstd-3.2.htx>`_, which is not
        terribly precise.  The exact grammar is here is based on the
        YACC grammar in the `unity library <https://purl.org/nxg/dist/unity/>`_.
        """
        tokens = cls._tokens

        def p_main(p):
            """
            main : factor combined_units
                 | combined_units
                 | DIMENSIONLESS
                 | OPEN_BRACKET combined_units CLOSE_BRACKET
                 | OPEN_BRACKET DIMENSIONLESS CLOSE_BRACKET
                 | factor
            """
            from astropy.units import dex

            if len(p) == 3:
                p[0] = CompositeUnit(p[1] * p[2].scale, p[2].bases, p[2].powers)
            elif len(p) == 4:
                p[0] = dex(p[2])
            else:
                p[0] = Unit(p[1])

        def p_combined_units(p):
            """
            combined_units : product_of_units
                           | division_of_units
            """
            p[0] = p[1]

        def p_product_of_units(p):
            """
            product_of_units : unit_expression PRODUCT combined_units
                             | unit_expression
            """
            if len(p) == 4:
                p[0] = p[1] * p[3]
            else:
                p[0] = p[1]

        def p_division_of_units(p):
            """
            division_of_units : DIVISION unit_expression
                              | combined_units DIVISION unit_expression
            """
            if len(p) == 3:
                p[0] = p[2] ** -1
            else:
                p[0] = p[1] / p[3]

        def p_unit_expression(p):
            """
            unit_expression : unit_with_power
                            | OPEN_PAREN combined_units CLOSE_PAREN
            """
            if len(p) == 2:
                p[0] = p[1]
            else:
                p[0] = p[2]

        def p_factor(p):
            """
            factor : signed_float X UINT SIGN UINT
                   | UINT X UINT SIGN UINT
                   | UINT SIGN UINT
                   | UINT
                   | signed_float
            """
            match p[1:]:
                case factor, _, 10, sign, exponent:
                    p[0] = factor * 10.0 ** (sign * exponent)
                case _, _, _, _, _:
                    raise ValueError("Only base ten exponents are allowed in CDS")
                case 10, sign, exponent:
                    p[0] = 10.0 ** (sign * exponent)
                case _, _, _:
                    raise ValueError("Only base ten exponents are allowed in CDS")
                case _:
                    p[0] = p[1]

        def p_unit_with_power(p):
            """
            unit_with_power : UNIT numeric_power
                            | UNIT
            """
            if len(p) == 2:
                p[0] = p[1]
            else:
                p[0] = p[1] ** p[2]

        def p_numeric_power(p):
            """
            numeric_power : sign UINT
            """
            p[0] = p[1] * p[2]

        def p_sign(p):
            """
            sign : SIGN
                 |
            """
            if len(p) == 2:
                p[0] = p[1]
            else:
                p[0] = 1.0

        def p_signed_float(p):
            """
            signed_float : sign UINT
                         | sign UFLOAT
            """
            p[0] = p[1] * p[2]

        def p_error(p):
            raise ValueError()

        return parsing.yacc(tabmodule="cds_parsetab", package="astropy/units")

    @classmethod
    def parse(cls, s: str, debug: bool = False) -> UnitBase:
        if " " in s:
            raise ValueError("CDS unit must not contain whitespace")
        if not isinstance(s, str):
            s = s.decode("ascii")

        return cls._do_parse(s, debug)

    @classmethod
    def _format_mantissa(cls, m: str) -> str:
        return "" if m == "1" else m

    @classmethod
    def _format_superscript(cls, number: str) -> str:
        return number if number.startswith("-") else "+" + number

    @classmethod
    def to_string(
        cls,
        unit: UnitBase,
        fraction: bool | Literal["inline", "multiline"] = False,
        deprecations: DeprecatedUnitAction = DeprecatedUnitAction.WARN,
    ) -> str:
        # Remove units that aren't known to the format
        unit = cls._decompose_to_known_units(unit)

        if not unit.bases:
            if unit.scale == 1:
                return "---"
            elif is_effectively_unity(unit.scale * 100.0):
                return "%"

        return super().to_string(unit, fraction=fraction)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
Functions to do XML schema and DTD validation.  At the moment, this
makes a subprocess call to xmllint.  This could use a Python-based
library at some point in the future, if something appropriate could be
found.
"""

import os
import subprocess
from signal import Signals


def validate_schema(filename, schema_file):
    """
    Validates an XML file against a schema or DTD.

    Parameters
    ----------
    filename : str
        The path to the XML file to validate

    schema_file : str
        The path to the XML schema or DTD

    Returns
    -------
    returncode, stdout, stderr : int, str, str
        Returns the returncode from xmllint and the stdout and stderr
        as strings
    """
    base, ext = os.path.splitext(schema_file)
    if ext == ".xsd":
        schema_part = "--schema"
    elif ext == ".dtd":
        schema_part = "--dtdvalid"
    else:
        raise TypeError("schema_file must be a path to an XML Schema or DTD")

    p = subprocess.Popen(
        ["xmllint", "--noout", "--nonet", schema_part, schema_file, filename],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    stdout, stderr = p.communicate()

    if p.returncode == 127:
        raise OSError("xmllint not found, so can not validate schema")
    elif p.returncode < 0:
        raise OSError(
            f"xmllint was terminated by signal '{Signals(-p.returncode).name}'"
        )

    return p.returncode, stdout, stderr



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This package defines the SI units.  They are also available in
(and should be used through) the `astropy.units` namespace.
"""
# avoid ruff complaints about undefined names defined by def_unit
# ruff: noqa: F821

import numpy as np

from .core import CompositeUnit, def_unit
from .docgen import generate_dunder_all, generate_unit_summary

__all__: list[str] = []  #  Units are added at the end

_ns = globals()


###########################################################################
# DIMENSIONLESS

def_unit(
    ["percent", "pct"],
    CompositeUnit(0.01, [], []),
    namespace=_ns,
    prefixes=False,
    doc="percent: one hundredth of unity, factor 0.01",
    format={"generic": "%", "console": "%", "cds": "%", "latex": r"\%", "unicode": "%"},
)

###########################################################################
# LENGTH

# We exclude c as a prefix so we can define cm as a derived unit,
# not a PrefixUnit. That way it can be used in cgs as a base unit.
def_unit(
    ["m", "meter"],
    namespace=_ns,
    prefixes=True,
    exclude_prefixes=["c"],
    doc="meter: base unit of length in SI",
)
def_unit(
    ["cm", "centimeter"],
    0.01 * m,
    namespace=_ns,
    prefixes=False,
    doc="cm: 0.01 m in SI, base unit of length in cgs",
)
def_unit(
    ["micron"],
    um,
    namespace=_ns,
    doc="micron: alias for micrometer (um)",
    format={"latex": r"\mu m", "unicode": "\N{MICRO SIGN}m"},
)
def_unit(
    ["Angstrom", "AA", "angstrom", "Å"],
    0.1 * nm,
    namespace=_ns,
    doc="ångström: 10 ** -10 m",
    prefixes=[(["m", "milli"], ["milli", "m"], 1.0e-3)],
    format={
        "latex": r"\mathring{A}",
        "ogip": "angstrom",
        "unicode": "Å",
        "vounit": "Angstrom",
    },
)


###########################################################################
# AREA

def_unit(
    ["ha", "hectare"],
    1e4 * m**2,
    namespace=_ns,
    prefixes=False,
    doc="hectare: unit of area, used to express area of land",
)


###########################################################################
# VOLUMES

def_unit(
    (["l", "L", "ℓ"], ["liter"]),
    1000 * cm**3.0,
    namespace=_ns,
    prefixes=True,
    doc="liter: metric unit of volume",
)


###########################################################################
# ANGULAR MEASUREMENTS

def_unit(
    ["rad", "radian"],
    namespace=_ns,
    prefixes=True,
    doc=(
        "radian: angular measurement of the ratio between the length "
        "on an arc and its radius"
    ),
)
def_unit(
    ["deg", "degree"],
    np.pi / 180.0 * rad,
    namespace=_ns,
    prefixes=True,
    doc="degree: angular measurement 1/360 of full rotation",
)
def_unit(
    ["hourangle"],
    15.0 * deg,
    namespace=_ns,
    prefixes=False,
    doc="hour angle: angular measurement with 24 in a full circle",
    format={"latex": r"{}^{h}", "unicode": "ʰ"},
)
def_unit(
    ["arcmin", "arcminute"],
    1.0 / 60.0 * deg,
    namespace=_ns,
    prefixes=True,
    doc="arc minute: angular measurement",
)
def_unit(
    ["arcsec", "arcsecond"],
    1.0 / 3600.0 * deg,
    namespace=_ns,
    prefixes=True,
    doc="arc second: angular measurement",
)
# These special formats should only be used for the non-prefix versions
deg._format = {"latex": r"{}^{\circ}", "unicode": "°"}
arcmin._format = {"latex": r"{}^{\prime}", "unicode": "′"}
arcsec._format = {"latex": r"{}^{\prime\prime}", "unicode": "″"}
def_unit(
    ["mas"],
    0.001 * arcsec,
    namespace=_ns,
    doc="milli arc second: angular measurement",
)
def_unit(
    ["uas", "\N{MICRO SIGN}as", "\N{GREEK SMALL LETTER MU}as"],
    0.000001 * arcsec,
    namespace=_ns,
    doc="micro arc second: angular measurement",
    format={"latex": r"\mu as", "unicode": "μas"},
)
def_unit(
    ["sr", "steradian"],
    rad**2,
    namespace=_ns,
    prefixes=True,
    doc="steradian: unit of solid angle in SI",
)


###########################################################################
# TIME

def_unit(
    ["s", "second"],
    namespace=_ns,
    prefixes=True,
    doc="second: base unit of time in SI.",
)
def_unit(
    ["min", "minute"],
    60 * s,
    prefixes=True,
    namespace=_ns,
)
def_unit(
    ["h", "hour", "hr"],
    3600 * s,
    namespace=_ns,
    prefixes=True,
    exclude_prefixes=["p"],
)
def_unit(
    ["d", "day"],
    24 * h,
    namespace=_ns,
    prefixes=True,
    exclude_prefixes=["c", "y"],
)
def_unit(
    ["sday"],
    86164.09053 * s,
    namespace=_ns,
    doc="Sidereal day (sday) is the time of one rotation of the Earth.",
)
def_unit(
    ["wk", "week"],
    7 * day,
    namespace=_ns,
)
def_unit(
    ["fortnight"],
    2 * wk,
    namespace=_ns,
)
def_unit(
    ["a", "annum"],
    365.25 * d,
    namespace=_ns,
    prefixes=True,
    exclude_prefixes=["P", "h"],  # Avoid possible confusion with Pascal and hectare
)
def_unit(
    ["yr", "year"],
    365.25 * d,
    namespace=_ns,
    prefixes=True,
)


###########################################################################
# FREQUENCY

def_unit(
    ["Hz", "Hertz", "hertz"],
    1 / s,
    namespace=_ns,
    prefixes=True,
    doc="Frequency",
)


###########################################################################
# MASS

def_unit(
    ["kg", "kilogram"],
    namespace=_ns,
    doc="kilogram: base unit of mass in SI.",
)
def_unit(
    ["g", "gram"],
    1.0e-3 * kg,
    namespace=_ns,
    prefixes=True,
    exclude_prefixes=["k", "kilo"],
)
def_unit(
    ["t", "tonne"],
    1000 * kg,
    namespace=_ns,
    doc="Metric tonne",
)


###########################################################################
# AMOUNT OF SUBSTANCE

def_unit(
    ["mol", "mole"],
    namespace=_ns,
    prefixes=True,
    doc="mole: amount of a chemical substance in SI.",
)
def_unit(
    ["kat", "katal"],
    mol * s**-1,
    namespace=_ns,
    prefixes=True,
    doc="katal: catalytic activity.",
)


###########################################################################
# TEMPERATURE

def_unit(
    ["K", "Kelvin"],
    namespace=_ns,
    prefixes=True,
    doc="Kelvin: temperature with a null point at absolute zero.",
)
def_unit(
    ["deg_C", "Celsius"],
    namespace=_ns,
    doc="Degrees Celsius",
    format={"latex": r"{}^{\circ}C", "unicode": "°C", "fits": "Celsius"},
)


###########################################################################
# FORCE

def_unit(
    ["N", "Newton", "newton"],
    kg * m * s**-2,
    namespace=_ns,
    prefixes=True,
    doc="Newton: force",
)


##########################################################################
# ENERGY

def_unit(
    ["J", "Joule", "joule"],
    N * m,
    namespace=_ns,
    prefixes=True,
    doc="Joule: energy",
)


##########################################################################
# PRESSURE

def_unit(
    ["Pa", "Pascal", "pascal"],
    J * m**-3,
    namespace=_ns,
    prefixes=True,
    doc="Pascal: pressure",
)


###########################################################################
# POWER

def_unit(
    ["W", "Watt", "watt"],
    J / s,
    namespace=_ns,
    prefixes=True,
    doc="Watt: power",
)


###########################################################################
# ELECTRICAL

def_unit(
    ["A", "ampere", "amp"],
    namespace=_ns,
    prefixes=True,
    doc="ampere: base unit of electric current in SI",
)
def_unit(
    ["C", "coulomb"],
    A * s,
    namespace=_ns,
    prefixes=True,
    doc="coulomb: electric charge",
)
def_unit(
    ["V", "Volt", "volt"],
    J * C**-1,
    namespace=_ns,
    prefixes=True,
    doc="Volt: electric potential or electromotive force",
)
def_unit(
    (["Ohm", "ohm", "Ω"], ["Ohm"]),
    V * A**-1,
    namespace=_ns,
    prefixes=True,
    doc="Ohm: electrical resistance",
    format={"latex": r"\Omega", "ogip": "ohm", "unicode": "Ω"},
)
def_unit(
    ["S", "Siemens", "siemens"],
    A * V**-1,
    namespace=_ns,
    prefixes=True,
    doc="Siemens: electrical conductance",
)
def_unit(
    ["F", "Farad", "farad"],
    C * V**-1,
    namespace=_ns,
    prefixes=True,
    doc="Farad: electrical capacitance",
)


###########################################################################
# MAGNETIC

def_unit(
    ["Wb", "Weber", "weber"],
    V * s,
    namespace=_ns,
    prefixes=True,
    doc="Weber: magnetic flux",
)
def_unit(
    ["T", "Tesla", "tesla"],
    Wb * m**-2,
    namespace=_ns,
    prefixes=True,
    doc="Tesla: magnetic flux density",
)
def_unit(
    ["H", "Henry", "henry"],
    Wb * A**-1,
    namespace=_ns,
    prefixes=True,
    doc="Henry: inductance",
)


###########################################################################
# ILLUMINATION

def_unit(
    ["cd", "candela"],
    namespace=_ns,
    prefixes=True,
    doc="candela: base unit of luminous intensity in SI",
)
def_unit(
    ["lm", "lumen"],
    cd * sr,
    namespace=_ns,
    prefixes=True,
    doc="lumen: luminous flux",
)
def_unit(
    ["lx", "lux"],
    lm * m**-2,
    namespace=_ns,
    prefixes=True,
    doc="lux: luminous emittance",
)

###########################################################################
# RADIOACTIVITY

def_unit(
    ["Bq", "becquerel"],
    1 / s,
    namespace=_ns,
    prefixes=True,
    doc="becquerel: unit of radioactivity",
)
def_unit(
    ["Ci", "curie"],
    Bq * 3.7e10,
    namespace=_ns,
    prefixes=True,
    doc="curie: unit of radioactivity",
)
def_unit(
    ["Gy", "gray"],
    J * kg**-1,
    namespace=_ns,
    prefixes=True,
    doc="gray: absorbed dose of ionizing radiation or kinetic energy released per unit mass (kerma)",
)
def_unit(
    ["Sv", "sievert"],
    J * kg**-1,
    namespace=_ns,
    prefixes=True,
    doc="sievert: equivalent dose of ionizing radiation",
)


###########################################################################
# BASES

bases = {m, s, kg, A, cd, rad, K, mol}


###########################################################################
# ALL & DOCSTRING

__all__ += generate_dunder_all(globals())  # noqa: PLE0605

if __doc__ is not None:
    # This generates a docstring for this module that describes all of the
    # standard units defined here.
    __doc__ += generate_unit_summary(globals())



import os
import re

import numpy as np
import pytest

from astropy.table.scripts import showtable
from astropy.units import UnitsWarning

ROOT = os.path.abspath(os.path.dirname(__file__))
ASCII_ROOT = os.path.join(ROOT, "..", "..", "io", "ascii", "tests")
FITS_ROOT = os.path.join(ROOT, "..", "..", "io", "fits", "tests")
VOTABLE_ROOT = os.path.join(ROOT, "..", "..", "io", "votable", "tests")


def test_missing_file(capsys):
    showtable.main(["foobar.fits"])
    out, err = capsys.readouterr()
    assert err.startswith("ERROR: [Errno 2] No such file or directory: 'foobar.fits'")


def test_info(capsys):
    showtable.main([os.path.join(FITS_ROOT, "data/table.fits"), "--info"])
    out, err = capsys.readouterr()
    assert out.splitlines() == [
        "<Table length=3>",
        " name   dtype ",
        "------ -------",
        "target bytes20",
        " V_mag float32",
    ]


def test_stats(capsys):
    showtable.main([os.path.join(FITS_ROOT, "data/table.fits"), "--stats"])
    out, err = capsys.readouterr()
    expected = [
        "<Table length=3>",
        " name    mean    std   min  max ",
        "------ ------- ------- ---- ----",
        "target      --      --   --   --",
        " V_mag 12.866[0-9]? 1.72111 11.1 15.2",
    ]

    out = out.splitlines()
    assert out[:4] == expected[:4]
    # Here we use re.match as in some cases one of the values above is
    # platform-dependent.
    assert re.match(expected[4], out[4]) is not None


def test_fits(capsys):
    showtable.main([os.path.join(FITS_ROOT, "data/table.fits")])
    out, err = capsys.readouterr()
    assert out.splitlines() == [
        " target V_mag",
        "------- -----",
        "NGC1001  11.1",
        "NGC1002  12.3",
        "NGC1003  15.2",
    ]


def test_fits_hdu(capsys):
    with pytest.warns(UnitsWarning):
        showtable.main(
            [
                os.path.join(FITS_ROOT, "data/zerowidth.fits"),
                "--hdu",
                "AIPS OF",
            ]
        )

    out, err = capsys.readouterr()
    assert out.startswith(
        "   TIME    SOURCE ID ANTENNA NO. SUBARRAY FREQ ID ANT FLAG STATUS 1\n"
        "   DAYS                                                            \n"
        "---------- --------- ----------- -------- ------- -------- --------\n"
        "0.14438657         1          10        1       1        4        4\n"
    )


def test_csv(capsys):
    showtable.main([os.path.join(ASCII_ROOT, "data/simple_csv.csv")])
    out, err = capsys.readouterr()
    assert out.splitlines() == [
        " a   b   c ",
        "--- --- ---",
        "  1   2   3",
        "  4   5   6",
    ]


def test_ascii_format(capsys):
    showtable.main(
        [
            os.path.join(ASCII_ROOT, "data/commented_header.dat"),
            "--format",
            "ascii.commented_header",
        ]
    )
    out, err = capsys.readouterr()
    assert out.splitlines() == [
        " a   b   c ",
        "--- --- ---",
        "  1   2   3",
        "  4   5   6",
    ]


def test_ascii_delimiter(capsys):
    showtable.main(
        [
            os.path.join(ASCII_ROOT, "data/simple2.txt"),
            "--format",
            "ascii",
            "--delimiter",
            "|",
        ]
    )
    out, err = capsys.readouterr()
    assert out.splitlines() == [
        "obsid redshift  X    Y      object   rad ",
        "----- -------- ---- ---- ----------- ----",
        " 3102     0.32 4167 4085 Q1250+568-A  9.0",
        " 3102     0.32 4706 3916 Q1250+568-B 14.0",
        "  877     0.22 4378 3892 'Source 82' 12.5",
    ]


def test_votable(capsys):
    with np.errstate(over="ignore"):
        # https://github.com/astropy/astropy/issues/13341
        showtable.main(
            [
                os.path.join(VOTABLE_ROOT, "data/regression.xml"),
                "--table-id",
                "main_table",
                "--max-width",
                "50",
            ]
        )
    out, err = capsys.readouterr()
    assert out.splitlines() == [
        "   string_test    string_test_2 ...   bitarray2  ",
        "----------------- ------------- ... -------------",
        "    String & test    Fixed stri ... True .. False",
        "String &amp; test    0123456789 ...      -- .. --",
        "             XXXX          XXXX ...      -- .. --",
        "                                ...      -- .. --",
        "                                ...      -- .. --",
    ]


def test_max_lines(capsys):
    showtable.main(
        [
            os.path.join(ASCII_ROOT, "data/cds2.dat"),
            "--format",
            "ascii.cds",
            "--max-lines",
            "7",
            "--max-width",
            "30",
        ]
    )
    out, err = capsys.readouterr()
    assert out.splitlines() == [
        "      SST       ... Note",
        "                ...     ",
        "--------------- ... ----",
        "041314.1+281910 ...   --",
        "            ... ...  ...",
        "044427.1+251216 ...   --",
        "044642.6+245903 ...   --",
        "Length = 215 rows",
    ]


def test_show_dtype(capsys):
    showtable.main([os.path.join(FITS_ROOT, "data/table.fits"), "--show-dtype"])
    out, err = capsys.readouterr()
    assert out.splitlines() == [
        " target  V_mag ",
        "bytes20 float32",
        "------- -------",
        "NGC1001    11.1",
        "NGC1002    12.3",
        "NGC1003    15.2",
    ]


def test_hide_unit(capsys):
    showtable.main([os.path.join(ASCII_ROOT, "data/cds.dat"), "--format", "ascii.cds"])
    out, err = capsys.readouterr()
    assert out.splitlines() == [
        "Index RAh RAm  RAs  DE- DEd  DEm    DEs   Match Class  AK   Fit   ",
        "       h  min   s       deg arcmin arcsec             mag GsolMass",
        "----- --- --- ----- --- --- ------ ------ ----- ----- --- --------",
        "    1   3  28 39.09   +  31      6    1.9    --    I*  --     1.35",
    ]

    showtable.main(
        [
            os.path.join(ASCII_ROOT, "data/cds.dat"),
            "--format",
            "ascii.cds",
            "--hide-unit",
        ]
    )
    out, err = capsys.readouterr()
    assert out.splitlines() == [
        "Index RAh RAm  RAs  DE- DEd DEm DEs Match Class  AK Fit ",
        "----- --- --- ----- --- --- --- --- ----- ----- --- ----",
        "    1   3  28 39.09   +  31   6 1.9    --    I*  -- 1.35",
    ]
